#include <stack>
#include <time.h>
#include <iomanip>

#include "ViewCell.h"
#include "Plane3.h"
#include "VspTree.h"
#include "Mesh.h"
#include "common.h"
#include "Environment.h"
#include "Polygon3.h"
#include "Ray.h"
#include "AxisAlignedBox3.h"
#include "Exporter.h"
#include "Plane3.h"
#include "ViewCellsManager.h"
#include "Beam.h"
#include "KdTree.h"
#include "IntersectableWrapper.h"
#include "HierarchyManager.h"
#include "BvHierarchy.h"
#include "OspTree.h"



namespace GtpVisibilityPreprocessor {


#define VISUALIZE_SPLIT 0
#define USE_FIXEDPOINT_T 0
#define HACK_PERFORMANCE 1


#define TYPE_INTERIOR -2
#define TYPE_LEAF -3


/////////////
//-- static members

VspTree *VspTree::VspSubdivisionCandidate::sVspTree = NULL;
int VspNode::sMailId = 1;

// variable for debugging volume contribution for heuristics
static float debugVol;



inline static bool ilt(Intersectable *obj1, Intersectable *obj2)
{
	return obj1->mId < obj2->mId;
}


// pvs penalty can be different from pvs size
inline static float EvalPvsPenalty(const float pvs, 
								   const float lower,
								   const float upper)
{
	// clamp to minmax values
	if (pvs < lower)
	{
		return (float)lower;
	}
	else if (pvs > upper)
	{
		return (float)upper;
	}
	return (float)pvs;
}


void VspTreeStatistics::Print(ostream &app) const
{
	app << "=========== VspTree statistics ===============\n";

	app << setprecision(4);

	app << "#N_CTIME  ( Construction time [s] )\n" << Time() << " \n";

	app << "#N_NODES ( Number of nodes )\n" << nodes << "\n";

	app << "#N_INTERIORS ( Number of interior nodes )\n" << Interior() << "\n";

	app << "#N_LEAVES ( Number of leaves )\n" << Leaves() << "\n";

	app << "#N_SPLITS ( Number of splits in axes x y z)\n";

	for (int i = 0; i < 3; ++ i)
		app << splits[i] << " ";

	app << endl;

	app << "#N_PMAXDEPTHLEAVES ( Percentage of leaves at maximum depth )\n" 
		<<	maxDepthNodes * 100 / (double)Leaves() << endl;

	app << "#N_PMINPVSLEAVES  ( Percentage of leaves with mininimal PVS )\n" 
		<< minPvsNodes * 100 / (double)Leaves() << endl;

	app << "#N_PMINRAYSLEAVES  ( Percentage of leaves with minimal number of rays)\n" 
		<< minRaysNodes * 100 / (double)Leaves() << endl;

	app << "#N_MAXCOSTNODES  ( Percentage of leaves with terminated because of max cost ratio )\n"
		<< maxCostNodes * 100 / (double)Leaves() << endl;

	app << "#N_PMINPROBABILITYLEAVES  ( Percentage of leaves with mininum probability )\n"
		<< minProbabilityNodes * 100 / (double)Leaves() << endl;

	app << "#N_PMAXRAYCONTRIBLEAVES  ( Percentage of leaves with maximal ray contribution )\n" 
		<<	maxRayContribNodes * 100 / (double)Leaves() << endl;

	app << "#N_PMAXDEPTH ( Maximal reached depth )\n" << maxDepth << endl;

	app << "#N_PMINDEPTH ( Minimal reached depth )\n" << minDepth << endl;

	app << "#AVGDEPTH ( average depth )\n" << AvgDepth() << endl;

	app << "#N_INVALIDLEAVES (number of invalid leaves )\n" << invalidLeaves << endl;

	app << "#AVGRAYS (number of rays / leaf)\n" << AvgRays() << endl;
	
	app << "#N_GLOBALCOSTMISSES ( Global cost misses )\n" << mGlobalCostMisses << endl;

	app << "========== END OF VspTree statistics ==========\n";
}



VspTree::VspTraversalData::VspTraversalData():
mNode(NULL),
mDepth(0),
mRays(NULL),
mPvs(0),
mProbability(0.0),
mMaxCostMisses(0), 
mPriority(0),
mCorrectedPvs(0)
{}


VspTree::VspTraversalData::VspTraversalData(VspLeaf *node, 
											const int depth, 
											RayInfoContainer *rays,
											const float pvs,
											const float p,
											const AxisAlignedBox3 &box):
mNode(node), 
mDepth(depth), 
mRays(rays),
mProbability(p),
mBoundingBox(box),
mMaxCostMisses(0),
mPriority(0),
mCorrectedPvs(0),
mPvs(pvs),
mRenderCost(0),
mCorrectedRenderCost(0)
{}


float VspTree::VspTraversalData::GetAvgRayContribution() const
{
	return (float)mPvs / ((float)mRays->size() + Limits::Small);
}


float VspTree::VspTraversalData::GetAvgRaysPerObject() const
{
	return (float)mRays->size() / ((float)mPvs + Limits::Small);
}

	
float VspTree::VspTraversalData::GetCost() const 
{
	return mPriority; 
}
	

void VspTree::VspTraversalData::Clear()
{
	DEL_PTR(mRays);

	if (mNode)
	{	// delete old view cell
		delete mNode->GetViewCell();
		delete mNode;
		mNode = NULL;
	}
}


/******************************************************************/
/*                  class VspNode implementation                  */
/******************************************************************/


VspNode::VspNode(): 
mParent(NULL), 
mTreeValid(true), 
mTimeStamp(0)
{}


VspNode::VspNode(VspInterior *parent): 
mParent(parent), 
mTreeValid(true), 
mTimeStamp(0)
{}


bool VspNode::IsRoot() const 
{
	return mParent == NULL;
}


VspInterior *VspNode::GetParent() 
{ 
	return mParent; 
}


void VspNode::SetParent(VspInterior *parent)
{
	mParent = parent;
}


bool VspNode::IsSibling(VspNode *n) const
{
	return  ((this != n) && mParent && 
			 (mParent->GetFront() == n) || (mParent->GetBack() == n));
}


int VspNode::GetDepth() const
{
	int depth = 0;
	VspNode *p = mParent;
	
	while (p)
	{
		p = p->mParent;
		++ depth;
	}

	return depth;
}


bool VspNode::TreeValid() const
{
	return mTreeValid;
}


void VspNode::SetTreeValid(const bool v)
{
	mTreeValid = v;
}



/****************************************************************/
/*              class VspInterior implementation                */
/****************************************************************/


VspInterior::VspInterior(const AxisAlignedPlane &plane): 
mPlane(plane), mFront(NULL), mBack(NULL)
{}


VspInterior::~VspInterior() 
{ 
	DEL_PTR(mFront); 
	DEL_PTR(mBack);
}






void VspInterior::ReplaceChildLink(VspNode *oldChild, VspNode *newChild) 
{
	if (mBack == oldChild)
		mBack = newChild;
	else
		mFront = newChild;
}


void VspInterior::SetupChildLinks(VspNode *front, VspNode *back) 
{
    mBack = back;
    mFront = front;
}


AxisAlignedBox3 VspInterior::GetBoundingBox() const
{
	return mBoundingBox;
}


void VspInterior::SetBoundingBox(const AxisAlignedBox3 &box)
{
	mBoundingBox = box;
}


int VspInterior::Type() const
{
	return Interior;
}



/****************************************************************/
/*                  class VspLeaf implementation                */
/****************************************************************/


VspLeaf::VspLeaf(): 
mViewCell(NULL), mSubdivisionCandidate(NULL)
{
}


VspLeaf::~VspLeaf()
{
	VssRayContainer::const_iterator vit, vit_end = mVssRays.end();

	for (vit = mVssRays.begin(); vit != vit_end; ++ vit)
	{
		VssRay *ray = *vit;
		ray->Unref();

		if (!ray->IsActive())
			delete ray;
	}
}


int VspLeaf::Type() const
{
	return Leaf;
}


VspLeaf::VspLeaf(ViewCellLeaf *viewCell): 
mViewCell(viewCell)
{
}


VspLeaf::VspLeaf(VspInterior *parent): 
VspNode(parent), mViewCell(NULL)
{}


VspLeaf::VspLeaf(VspInterior *parent, ViewCellLeaf *viewCell): 
VspNode(parent), mViewCell(viewCell)
{
}




void VspLeaf::SetViewCell(ViewCellLeaf *viewCell)
{
	mViewCell = viewCell;
}





/*************************************************************************/
/*                       class VspTree implementation                    */
/*************************************************************************/


VspTree::VspTree():
mRoot(NULL),
mOutOfBoundsCell(NULL),
mStoreRays(false),
//mStoreRays(true),
mTimeStamp(1),
mHierarchyManager(NULL)
{
	mLocalSubdivisionCandidates = new vector<SortableEntry>;

	bool randomize = false;
	Environment::GetSingleton()->GetBoolValue("VspTree.Construction.randomize", randomize);
	if (randomize)
		Randomize(); // initialise random generator for heuristics

	char subdivisionStatsLog[100];
	Environment::GetSingleton()->GetStringValue("VspTree.subdivisionStats", subdivisionStatsLog);
	mSubdivisionStats.open(subdivisionStatsLog);

	/////////////
	//-- termination criteria for autopartition

	Environment::GetSingleton()->GetIntValue("VspTree.Termination.maxDepth", mTermMaxDepth);
	Environment::GetSingleton()->GetIntValue("VspTree.Termination.minPvs", mTermMinPvs);
	Environment::GetSingleton()->GetIntValue("VspTree.Termination.minRays", mTermMinRays);
	Environment::GetSingleton()->GetFloatValue("VspTree.Termination.minProbability", mTermMinProbability);
	Environment::GetSingleton()->GetFloatValue("VspTree.Termination.maxRayContribution", mTermMaxRayContribution);
	
	Environment::GetSingleton()->GetIntValue("VspTree.Termination.missTolerance", mTermMissTolerance);
	Environment::GetSingleton()->GetIntValue("VspTree.Termination.maxViewCells", mMaxViewCells);
	// max cost ratio for early tree termination
	Environment::GetSingleton()->GetFloatValue("VspTree.Termination.maxCostRatio", mTermMaxCostRatio);

	Environment::GetSingleton()->GetFloatValue("VspTree.Termination.minGlobalCostRatio", mTermMinGlobalCostRatio);
	Environment::GetSingleton()->GetIntValue("VspTree.Termination.globalCostMissTolerance", mTermGlobalCostMissTolerance);

	Environment::GetSingleton()->GetFloatValue("VspTree.maxStaticMemory", mMaxMemory);


	//////////////
	//-- factors for bsp tree split plane heuristics

	Environment::GetSingleton()->GetFloatValue("VspTree.Termination.ct_div_ci", mCtDivCi);
	Environment::GetSingleton()->GetFloatValue("VspTree.Construction.epsilon", mEpsilon);
	Environment::GetSingleton()->GetFloatValue("VspTree.Construction.minBand", mMinBand);
	Environment::GetSingleton()->GetFloatValue("VspTree.Construction.maxBand", mMaxBand);
	Environment::GetSingleton()->GetIntValue("VspTree.maxTests", mMaxTests);

	Environment::GetSingleton()->GetFloatValue("VspTree.Construction.renderCostDecreaseWeight", mRenderCostDecreaseWeight);
	
	// if only the driving axis is used for axis aligned split
	Environment::GetSingleton()->GetBoolValue("VspTree.splitUseOnlyDrivingAxis", mOnlyDrivingAxis);
	Environment::GetSingleton()->GetBoolValue("VspTree.useCostHeuristics", mUseCostHeuristics);
	Environment::GetSingleton()->GetBoolValue("VspTree.simulateOctree", mCirculatingAxis);

	//mMemoryConst = (float)(sizeof(VspLeaf) + sizeof(VspViewCell));
	//mMemoryConst = 50;//(float)(sizeof(VspViewCell));
	//mMemoryConst = (float)(sizeof(VspLeaf) + sizeof(ObjectPvs));

	mMemoryConst = 16;//(float)sizeof(ObjectPvs);
	

	//////////////
	//-- debug output

	Debug << "******* VSP options ********" << endl;

    Debug << "max depth: " << mTermMaxDepth << endl;
	Debug << "min PVS: " << mTermMinPvs << endl;
	Debug << "min probabiliy: " << mTermMinProbability << endl;
	Debug << "min rays: " << mTermMinRays << endl;
	Debug << "max ray contri: " << mTermMaxRayContribution << endl;
	Debug << "max cost ratio: " << mTermMaxCostRatio << endl;
	Debug << "miss tolerance: " << mTermMissTolerance << endl;
	Debug << "max view cells: " << mMaxViewCells << endl;
	Debug << "randomize: " << randomize << endl;

	Debug << "min global cost ratio: " << mTermMinGlobalCostRatio << endl;
	Debug << "global cost miss tolerance: " << mTermGlobalCostMissTolerance << endl;
	Debug << "only driving axis: " << mOnlyDrivingAxis << endl;
	Debug << "max memory: " << mMaxMemory << endl;
	Debug << "use cost heuristics: " << mUseCostHeuristics << endl;
	Debug << "subdivision stats log: " << subdivisionStatsLog << endl;
	Debug << "render cost decrease weight: " << mRenderCostDecreaseWeight << endl;

	Debug << "circulating axis: " << mCirculatingAxis << endl;
	Debug << "minband: " << mMinBand << endl;
	Debug << "maxband: " << mMaxBand << endl;

	Debug << "vsp mem const: " << mMemoryConst << endl;

	Debug << endl;
}


VspViewCell *VspTree::GetOutOfBoundsCell()
{
	return mOutOfBoundsCell;
}


VspViewCell *VspTree::GetOrCreateOutOfBoundsCell()
{
	if (!mOutOfBoundsCell)
	{
		mOutOfBoundsCell = new VspViewCell();
		mOutOfBoundsCell->SetId(-1);
		mOutOfBoundsCell->SetValid(false);
	}

	return mOutOfBoundsCell;
}


const VspTreeStatistics &VspTree::GetStatistics() const
{
	return mVspStats;
}


VspTree::~VspTree()
{
	DEL_PTR(mRoot);
	DEL_PTR(mLocalSubdivisionCandidates);
}


void VspTree::ComputeBoundingBox(const VssRayContainer &rays,
								 AxisAlignedBox3 *forcedBoundingBox) 
{	
	if (forcedBoundingBox)
	{
		mBoundingBox = *forcedBoundingBox;
		return;
	}
	
	//////////////////////////////////////////////
	// bounding box of view space includes all visibility events

	mBoundingBox.Initialize();
	VssRayContainer::const_iterator rit, rit_end = rays.end();

	for (rit = rays.begin(); rit != rit_end; ++ rit)
	{
		VssRay *ray = *rit;

		mBoundingBox.Include(ray->GetTermination());
		mBoundingBox.Include(ray->GetOrigin());
	}
}


void VspTree::AddSubdivisionStats(const int viewCells,
								  const float renderCostDecr,
								  const float totalRenderCost,
								  const float avgRenderCost)
{
	mSubdivisionStats 
			<< "#ViewCells\n" << viewCells << endl
			<< "#RenderCostDecrease\n" << renderCostDecr << endl 
			<< "#TotalRenderCost\n" << totalRenderCost << endl
			<< "#AvgRenderCost\n" << avgRenderCost << endl;
}


// $$matt temporary: number of rayrefs + pvs should be in there, but
// commented out for testing
float VspTree::GetMemUsage() const
{
	return (float)
		 (sizeof(VspTree)
		  + mVspStats.Leaves() * sizeof(VspLeaf) 
		  + mVspStats.Leaves() * sizeof(VspViewCell)
		  + mVspStats.Interior() * sizeof(VspInterior)
		  //+ mVspStats.pvs * sizeof(PvsData)
		  //+ mVspStats.rayRefs * sizeof(RayInfo)
		  ) / (1024.0f * 1024.0f);
}


inline bool VspTree::LocalTerminationCriteriaMet(const VspTraversalData &data) const
{
	const bool localTerminationCriteriaMet = (0
		|| ((int)data.mRays->size() <= mTermMinRays)
		|| (data.mPvs <= mTermMinPvs)
		|| (data.mProbability <= mTermMinProbability)
		|| (data.mDepth >= mTermMaxDepth)
		);

#if GTP_DEBUG
	if (localTerminationCriteriaMet)
	{
		Debug << "local termination criteria met:" << endl;
		Debug << "rays: " << (int)data.mRays->size() << "  " << mTermMinRays << endl;
		Debug << "pvs: " << data.mPvs << " " << mTermMinPvs << endl;
		Debug << "p: " <<  data.mProbability << " " << mTermMinProbability << endl;
		Debug << "avg contri: " << data.GetAvgRayContribution() << " " << mTermMaxRayContribution << endl;
		Debug << "depth " << data.mDepth << " " << mTermMaxDepth << endl;
	}
#endif
	return localTerminationCriteriaMet;		
}


inline bool VspTree::GlobalTerminationCriteriaMet(const VspTraversalData &data) const
{
	// note: to track for global cost misses does not really 
	// make sense because termination on cost or memory is done 
	// in the hierarchy mananger
	const bool terminationCriteriaMet = (0
		// || mOutOfMemory
		|| (mVspStats.Leaves() >= mMaxViewCells)
		// || (mVspStats.mGlobalCostMisses >= mTermGlobalCostMissTolerance) 
		);

#if GTP_DEBUG
	if (terminationCriteriaMet)
	{
		Debug << "vsp global termination criteria met:" << endl;
		Debug << "cost misses: " << mVspStats.mGlobalCostMisses << " " << mTermGlobalCostMissTolerance << endl;
		Debug << "leaves: " << mVspStats.Leaves() << " " <<  mMaxViewCells << endl;
	}
#endif

	return terminationCriteriaMet;
}


void VspTree::CreateViewCell(VspTraversalData &tData, 
							 const bool updatePvs, 
							 const float renderCost,
							 const int pvs)
{
	///////////////
	//-- create new view cell

	VspLeaf *leaf = tData.mNode;

	VspViewCell *viewCell = new VspViewCell();
    leaf->SetViewCell(viewCell);
	
	int conSamp = 0;
	float sampCon = 0.0f;

	if (updatePvs)
	{
		// store pvs of view cell
		AddSamplesToPvs(leaf, *tData.mRays, sampCon, conSamp);

		// update scalar pvs size value
		ObjectPvs &pvs = viewCell->GetPvs();
		mViewCellsManager->UpdateScalarPvsSize(viewCell, 
											   pvs.EvalPvsCost(), 
											   pvs.GetSize());

		mVspStats.contributingSamples += conSamp;
		mVspStats.sampleContributions += (int)sampCon;
	}
	else
	{
		viewCell->SetTrianglesInPvs(renderCost);
		viewCell->SetEntriesInPvs(pvs);
		
		//cout << "create view cell with tri=" << (int)renderCost << " pvs=" << pvs << endl; 
	}

	if (mStoreRays)
	{
		///////////
		//-- store sampling rays

        RayInfoContainer::const_iterator rit, rit_end = tData.mRays->end();

		for (rit = tData.mRays->begin(); rit != rit_end; ++ rit)
		{
			VssRay *ray = new VssRay(*(*rit).mRay);
			ray->Ref();

			// note: should rather store rays with view cell
			leaf->mVssRays.push_back(ray);
		}
	}
		
	// set view cell values
	viewCell->mLeaves.push_back(leaf);

	viewCell->SetVolume(tData.mProbability);
    leaf->mProbability = tData.mProbability;
}


void VspTree::EvalSubdivisionStats(const SubdivisionCandidate &sc)
{
	const float costDecr = sc.GetRenderCostDecrease();
	
	AddSubdivisionStats(mVspStats.Leaves(),
						costDecr,
						mTotalCost,
						(float)mTotalPvsSize / (float)mVspStats.Leaves());
}


VspNode *VspTree::Subdivide(SplitQueue &tQueue,
							SubdivisionCandidate *splitCandidate,
							const bool globalCriteriaMet)
{
	mSubdivTimer.Entry();

	// todo remove dynamic cast
	VspSubdivisionCandidate *sc = 
		static_cast<VspSubdivisionCandidate *>(splitCandidate);

	VspTraversalData &tData = sc->mParentData;
	VspNode *newNode = tData.mNode;

	if (!LocalTerminationCriteriaMet(tData) && !globalCriteriaMet)
	{	
		///////////
		//-- continue subdivision

		VspTraversalData tFrontData;
		VspTraversalData tBackData;
		
		// create new interior node and two leaf node
		const int maxCostMisses = sc->GetMaxCostMisses();

		newNode = SubdivideNode(*sc, tFrontData, tBackData);

		// how often was max cost ratio missed in this branch?
		tFrontData.mMaxCostMisses = maxCostMisses;
		tBackData.mMaxCostMisses = maxCostMisses;
			
		mTotalCost -= sc->GetRenderCostDecrease();
		//mTotalPvsSize += (int)(tFrontData.mPvs + tBackData.mPvs - tData.mPvs);
		mPvsEntries += sc->GetPvsEntriesIncr();

		// subdivision statistics
		if (1) EvalSubdivisionStats(*sc);
		
		/////////////
		//-- evaluate new split candidates for global greedy cost heuristics

		VspSubdivisionCandidate *frontCandidate = new VspSubdivisionCandidate(tFrontData);
		VspSubdivisionCandidate *backCandidate = new VspSubdivisionCandidate(tBackData);

		EvalSubdivisionCandidate(*frontCandidate);
		EvalSubdivisionCandidate(*backCandidate);

		// cross reference
		tFrontData.mNode->SetSubdivisionCandidate(frontCandidate); 
		tBackData.mNode->SetSubdivisionCandidate(backCandidate);

		tQueue.Push(frontCandidate);
		tQueue.Push(backCandidate);

		// note: leaf is not destroyed because it is needed to collect 
		// dirty candidates in hierarchy manager
	}

	if (newNode->IsLeaf()) // subdivision terminated
	{
		VspLeaf *leaf = static_cast<VspLeaf *>(newNode);
	
		if (0 && mStoreRays)
		{
			//////////
			//-- store rays piercing this view cell

			RayInfoContainer::const_iterator it, it_end = tData.mRays->end();
			for (it = tData.mRays->begin(); it != it_end; ++ it)
			{
				(*it).mRay->Ref();			
				leaf->mVssRays.push_back((*it).mRay);
			}
		}

		// finally evaluate statistics for this leaf
		EvaluateLeafStats(tData);
		// detach subdivision candidate: this leaf is no candidate for
		// splitting anymore
		tData.mNode->SetSubdivisionCandidate(NULL); 
		// detach node so it won't get deleted
		tData.mNode = NULL;
	}

	mSubdivTimer.Exit();

	return newNode;
}


void VspTree::EvalSubdivisionCandidate(VspSubdivisionCandidate &splitCandidate, 
									   bool computeSplitPlane)
{
	mPlaneTimer.Entry();

	if (computeSplitPlane)
	{
		float frontProb;
		float backProb;

		// compute locally best split plane
		const float ratio = SelectSplitPlane(splitCandidate.mParentData, 
									 		 splitCandidate.mSplitPlane, 
											 frontProb, 
											 backProb);
	
		const bool maxCostRatioViolated = mTermMaxCostRatio < ratio;

		const int maxCostMisses = splitCandidate.mParentData.mMaxCostMisses;
		// max cost threshold violated?
		splitCandidate.SetMaxCostMisses(maxCostRatioViolated ? 
										maxCostMisses + 1: maxCostMisses);
	}
	
	mPlaneTimer.Exit();
	mEvalTimer.Entry();

	VspLeaf *leaf = static_cast<VspLeaf *>(splitCandidate.mParentData.mNode);


	//////////////
	//-- compute global decrease in render cost

	const AxisAlignedPlane candidatePlane = splitCandidate.mSplitPlane;

	RayInfoContainer::const_iterator rit, rit_end = splitCandidate.mParentData.mRays->end();

	SplitData sData;

	Intersectable::NewMail(3);
	KdLeaf::NewMail(3);

	for (rit = splitCandidate.mParentData.mRays->begin(); rit != rit_end; ++ rit)
	{
		RayInfo rayInf = *rit;

		float t;
		
		// classify ray
		const int cf = rayInf.ComputeRayIntersection(candidatePlane.mAxis, 
													 candidatePlane.mPosition, 
													 t);

		VssRay *ray = rayInf.mRay;

#if HACK_PERFORMANCE
		Intersectable *obj = (*ray).mTerminationObject;

		BvhLeaf *leaf = mBvHierarchy->GetLeaf(obj);

		// evaluate contribution of ray endpoint to front 
		// and back pvs with respect to the classification
		UpdateContributionsToPvs(leaf, cf, sData);

#if COUNT_ORIGIN_OBJECTS
		obj = (*ray).mOriginObject;

		if (obj)
		{
			leaf = mBvHierarchy->GetLeaf(obj);
			UpdateContributionsToPvs(leaf, cf, sData); 
		}
#endif

#else
		cerr << "TODO classify" << endl;
#endif
	}

	sData.mTotalRenderCost = mViewCellsManager->ComputeRenderCost(sData.mTotalTriangles, sData.mTotalObjects);
	sData.mBackRenderCost = mViewCellsManager->ComputeRenderCost(sData.mBackTriangles, sData.mBackObjects);
	sData.mFrontRenderCost = mViewCellsManager->ComputeRenderCost(sData.mFrontTriangles, sData.mFrontObjects);
	

	/////////////
	// avg ray contri

	const float avgRayContri = (float)sData.mTotalObjects / 
		((float)splitCandidate.mParentData.mRays->size() + Limits::Small);

	const float avgRaysPerObject = (float)splitCandidate.mParentData.mRays->size() / 
		((float)sData.mTotalObjects + Limits::Small);

	//cout << "avg rays per obj: " << avgRaysPerObject << endl;

	splitCandidate.SetAvgRayContribution(avgRayContri);
	splitCandidate.SetAvgRaysPerObject(avgRaysPerObject);

	// todo: compute old render cost in the function
	// using michi's render cost evaluation
	float oldRenderCost;
	const float renderCostDecr = EvalRenderCostDecrease(splitCandidate, oldRenderCost, sData);
	splitCandidate.SetRenderCostDecrease(renderCostDecr);

	// the increase in pvs entries num induced by this split
	const int pvsEntriesIncr = EvalPvsEntriesIncr(splitCandidate, sData);
	splitCandidate.SetPvsEntriesIncr(pvsEntriesIncr);

	splitCandidate.mFrontTriangles = (float)sData.mFrontTriangles;
	splitCandidate.mBackTriangles = (float)sData.mBackTriangles;
	
	// take render cost of node into account 
	// otherwise danger of being stuck in a local minimum!
	float priority = mRenderCostDecreaseWeight * renderCostDecr + (1.0f - mRenderCostDecreaseWeight) * oldRenderCost;

	if (mHierarchyManager->mConsiderMemory)
	{
		priority /= ((float)splitCandidate.GetPvsEntriesIncr() + mMemoryConst);
	}

	splitCandidate.SetPriority(priority);

	//cout << "vsp render cost decrease=" << renderCostDecr << endl;
	mEvalTimer.Exit();
}


int VspTree::EvalPvsEntriesIncr(VspSubdivisionCandidate &splitCandidate,
								const SplitData &sData) const
{
	const float oldPvsRatio = (splitCandidate.mParentData.mPvs > 0) ? 
		sData.mTotalObjects / splitCandidate.mParentData.mPvs : 1;
	const float correctedOldPvs = splitCandidate.mParentData.mCorrectedPvs * oldPvsRatio;

	splitCandidate.mCorrectedFrontPvs = 
		mHierarchyManager->EvalCorrectedPvs((float)sData.mFrontObjects, 
											(float)correctedOldPvs, 
											splitCandidate.GetAvgRaysPerObject());
	splitCandidate.mCorrectedBackPvs = 
		mHierarchyManager->EvalCorrectedPvs((float)sData.mBackObjects, 
											(float)correctedOldPvs, 
											splitCandidate.GetAvgRaysPerObject());
	
	splitCandidate.mFrontPvs = (float)sData.mFrontObjects;
	splitCandidate.mBackPvs = (float)sData.mBackObjects;

	return (int)(splitCandidate.mCorrectedFrontPvs + splitCandidate.mCorrectedBackPvs - correctedOldPvs);
}


VspInterior *VspTree::SubdivideNode(const VspSubdivisionCandidate &sc,
									VspTraversalData &frontData,
									VspTraversalData &backData)
{
	mNodeTimer.Entry();
	
	VspLeaf *leaf = static_cast<VspLeaf *>(sc.mParentData.mNode);

	const VspTraversalData &tData = sc.mParentData;
	const AxisAlignedPlane &splitPlane = sc.mSplitPlane;

	///////////////
	//-- new traversal values

	frontData.mDepth = tData.mDepth + 1;
	backData.mDepth = tData.mDepth + 1;

	frontData.mRays = new RayInfoContainer();
	backData.mRays = new RayInfoContainer();

	//-- subdivide rays
	SplitRays(splitPlane, *tData.mRays, *frontData.mRays, *backData.mRays);

	//-- compute pvs
	frontData.mPvs = sc.mFrontPvs;
	backData.mPvs = sc.mBackPvs;
	
	//-- compute pvs correction for coping with undersampling
	frontData.mCorrectedPvs = sc.mCorrectedFrontPvs;
	backData.mCorrectedPvs = sc.mCorrectedBackPvs;

	//-- split front and back node geometry and compute area
	tData.mBoundingBox.Split(splitPlane.mAxis, 
							 splitPlane.mPosition, 
							 frontData.mBoundingBox, 
							 backData.mBoundingBox);

	frontData.mProbability = frontData.mBoundingBox.GetVolume();
	backData.mProbability = tData.mProbability - frontData.mProbability;

	//-- compute render cost
	frontData.mRenderCost = sc.mFrontRenderCost;
	backData.mRenderCost = sc.mBackRenderCost;
	
	frontData.mCorrectedRenderCost = sc.mCorrectedFrontRenderCost;
	backData.mCorrectedRenderCost = sc.mCorrectedBackRenderCost;

	////////
	//-- update some stats
	
	if (tData.mDepth > mVspStats.maxDepth)
	{	
		mVspStats.maxDepth = tData.mDepth;
	}

	// two more leaves per split
	mVspStats.nodes += 2;
	// and a new split
	++ mVspStats.splits[splitPlane.mAxis];


	////////////////
	//-- create front and back and subdivide further

	VspInterior *interior = new VspInterior(splitPlane);
	VspInterior *parent = leaf->GetParent();

	// replace a link from node's parent
	if (parent)
	{
		parent->ReplaceChildLink(leaf, interior);
		interior->SetParent(parent);
	}
	else // new root
	{
		mRoot = interior;
	}

	VspLeaf *frontLeaf = new VspLeaf(interior);
	VspLeaf *backLeaf = new VspLeaf(interior);

	// and setup child links
	interior->SetupChildLinks(frontLeaf, backLeaf);
	
	// add bounding box
	interior->SetBoundingBox(tData.mBoundingBox);

	// set front and back leaf
	frontData.mNode = frontLeaf;
	backData.mNode = backLeaf;

	// create front and back view cell for the new leaves
	CreateViewCell(frontData, false, sc.mFrontTriangles, (int)sc.mFrontPvs);
	CreateViewCell(backData, false, sc.mBackTriangles, (int)sc.mBackPvs);

	// set the time stamp so the order of traversal can be reconstructed
	interior->mTimeStamp = mHierarchyManager->mTimeStamp ++;

	mNodeTimer.Exit();

	return interior;
}


void VspTree::AddSamplesToPvs(VspLeaf *leaf,
							  const RayInfoContainer &rays,
							  float &sampleContributions,
							  int &contributingSamples)
{
	sampleContributions = 0;
	contributingSamples = 0;
  
	RayInfoContainer::const_iterator it, it_end = rays.end();
  
	ViewCellLeaf *vc = leaf->GetViewCell();

	// add contributions from samples to the pvs
	for (it = rays.begin(); it != it_end; ++ it)
	{
		float sc = 0.0f;
		VssRay *ray = (*it).mRay;

		bool madeContrib = false;
		float contribution = 1;

		Intersectable *entry = 
			mHierarchyManager->GetIntersectable(ray->mTerminationObject, true);

		if (entry) 
		{
			madeContrib = 
				vc->GetPvs().AddSample(entry, ray->mPdf);

			sc += contribution;
		}
#if COUNT_ORIGIN_OBJECTS
		entry = mHierarchyManager->GetIntersectable(ray->mOriginObject, true);

		if (entry) 
		{
			madeContrib = 
				vc->GetPvs().AddSample(entry, ray->mPdf);

			sc += contribution;
		}
#endif
		if (madeContrib)
		{
			++ contributingSamples;
		}
	}
}


void VspTree::SortSubdivisionCandidates(const RayInfoContainer &rays,
										const int axis, 
										float minBand, 
										float maxBand)
{
	mSortTimer.Entry();

	mLocalSubdivisionCandidates->clear();

	const int requestedSize = 2 * (int)(rays.size());

	// creates a sorted split candidates array
	if (mLocalSubdivisionCandidates->capacity() > 500000 &&
		requestedSize < (int)(mLocalSubdivisionCandidates->capacity() / 10) )
	{
    	delete mLocalSubdivisionCandidates;
   		mLocalSubdivisionCandidates = new vector<SortableEntry>;
	}

	mLocalSubdivisionCandidates->reserve(requestedSize);

	float pos;
	RayInfoContainer::const_iterator rit, rit_end = rays.end();

	const bool delayMinEvent = false;

	////////////
	//-- insert all queries
	for (rit = rays.begin(); rit != rit_end; ++ rit)
	{
		const bool positive = (*rit).mRay->HasPosDir(axis);
		
		// origin point
		pos = (*rit).ExtrapOrigin(axis);
		const int oType = positive ? SortableEntry::ERayMin : SortableEntry::ERayMax;

		if (delayMinEvent && oType == SortableEntry::ERayMin)
			pos += mEpsilon; // could be useful feature for walls

		mLocalSubdivisionCandidates->push_back(SortableEntry(oType, pos, (*rit).mRay));

		// termination point
		pos = (*rit).ExtrapTermination(axis);
		const int tType = positive ? SortableEntry::ERayMax : SortableEntry::ERayMin;

		if (delayMinEvent && tType == SortableEntry::ERayMin)
			pos += mEpsilon; // could be useful feature for walls

		mLocalSubdivisionCandidates->push_back(SortableEntry(tType, pos, (*rit).mRay));
	}

	if (1)
		stable_sort(mLocalSubdivisionCandidates->begin(), mLocalSubdivisionCandidates->end());
	else
		sort(mLocalSubdivisionCandidates->begin(), mLocalSubdivisionCandidates->end());

	mSortTimer.Exit();
}


float VspTree::PrepareHeuristics(KdLeaf *leaf)
{	
	float pvsSize = 0;
	
	if (!leaf->Mailed())
	{
		leaf->Mail();
		leaf->mCounter = 1;
		// add objects without the objects which are in several kd leaves
		pvsSize += (int)(leaf->mObjects.size() - leaf->mMultipleObjects.size());
	}
	else
	{
		++ leaf->mCounter;
	}

	//-- the objects belonging to several leaves must be handled seperately
	ObjectContainer::const_iterator oit, oit_end = leaf->mMultipleObjects.end();

	for (oit = leaf->mMultipleObjects.begin(); oit != oit_end; ++ oit)
	{
		Intersectable *object = *oit;
						
		if (!object->Mailed())
		{
			object->Mail();
			object->mCounter = 1;

			++ pvsSize;
		}
		else
		{
			++ object->mCounter;
		}
	}
	
	return pvsSize;
}


float VspTree::PrepareHeuristics(const RayInfoContainer &rays)
{	
	Intersectable::NewMail();
	KdNode::NewMail();

	float pvsSize = 0;

	RayInfoContainer::const_iterator ri, ri_end = rays.end();

    // set all kd nodes / objects as belonging to the front pvs
	for (ri = rays.begin(); ri != ri_end; ++ ri)
	{
		VssRay *ray = (*ri).mRay;
		
		pvsSize += PrepareHeuristics(*ray, true);
#if COUNT_ORIGIN_OBJECTS
		pvsSize += PrepareHeuristics(*ray, false);
#endif
	}

	return pvsSize;
}


float VspTree::EvalMaxEventContribution(KdLeaf *leaf) const
{
	float pvs = 0;

	// leaf falls out of right pvs
	if (-- leaf->mCounter == 0)
	{
		pvs -= ((float)leaf->mObjects.size() - (float)leaf->mMultipleObjects.size());
	}

	//////
	//-- separately handle objects which are in several kd leaves

	ObjectContainer::const_iterator oit, oit_end = leaf->mMultipleObjects.end();

	for (oit = leaf->mMultipleObjects.begin(); oit != oit_end; ++ oit)
	{
		Intersectable *object = *oit;

		if (-- object->mCounter == 0)
		{
			++ pvs;
		}
	}

	return pvs;
}


float VspTree::EvalMinEventContribution(KdLeaf *leaf) const
{
	if (leaf->Mailed())
		return 0;
	
	leaf->Mail();

	// add objects without those which are part of several kd leaves
	float pvs = ((float)leaf->mObjects.size() - (float)leaf->mMultipleObjects.size());

	// separately handle objects which are part of several kd leaves 
	ObjectContainer::const_iterator oit, oit_end = leaf->mMultipleObjects.end();

	for (oit = leaf->mMultipleObjects.begin(); oit != oit_end; ++ oit)
	{
		Intersectable *object = *oit;

		// object not previously in pvs
		if (!object->Mailed())
		{
			object->Mail();
			++ pvs;
		}
	}	

	return pvs;
}


void VspTree::EvalHeuristics(const SortableEntry &ci,
							 float &pvsLeft,
							 float &pvsRight) const
{
	VssRay *ray = ci.ray;

	// eval changes in pvs causes by min event 
	if (ci.type == SortableEntry::ERayMin)
	{
		pvsLeft += EvalMinEventContribution(*ray, true);
#if COUNT_ORIGIN_OBJECTS
		pvsLeft += EvalMinEventContribution(*ray, false);
#endif
	}
	else // eval changes in pvs causes by max event 
	{
		pvsRight -= EvalMaxEventContribution(*ray, true);
#if COUNT_ORIGIN_OBJECTS
		pvsRight -= EvalMaxEventContribution(*ray, false);
#endif
	}
}


float VspTree::EvalLocalCostHeuristics(const VspTraversalData &tData,
									   const AxisAlignedBox3 &box,
									   const int axis,
									   float &position,
									   const RayInfoContainer &usedRays)
{
	const float minBox = box.Min(axis);
	const float maxBox = box.Max(axis);

	const float sizeBox = maxBox - minBox;

	const float minBand = minBox + mMinBand * sizeBox;
	const float maxBand = minBox + mMaxBand * sizeBox;

	// sort by the current dimension
	SortSubdivisionCandidates(usedRays, axis, minBand, maxBand);

	// prepare the sweep
	// note: returns pvs size => no need t give pvs size as function parameter
	const float pvsCost = PrepareHeuristics(usedRays);

	// go through the lists, count the number of objects left and right
	// and evaluate the following cost funcion:
	// C = ct_div_ci  + (ql*rl + qr*rr)/queries

	float pvsl = 0;
	float pvsr = pvsCost;

	float pvsBack = pvsl;
	float pvsFront = pvsr;

	float sum = pvsCost * sizeBox;
	float minSum = 1e20f;

	// if no good split can be found, take mid split
	position = minBox + 0.5f * sizeBox;
	
	// the relative cost ratio
	float ratio = 99999999.0f;
	bool splitPlaneFound = false;

	Intersectable::NewMail();
	KdLeaf::NewMail();

	const float volRatio = 
		tData.mBoundingBox.GetVolume() / (sizeBox * mBoundingBox.GetVolume());

	////////
	//-- iterate through visibility events

	vector<SortableEntry>::const_iterator ci, 
		ci_end = mLocalSubdivisionCandidates->end();

#ifdef GTP_DEBUG
	const int leaves = mVspStats.Leaves();
	const bool printStats = ((axis == 0) && (leaves > 0) && (leaves < 90));
	
	ofstream sumStats;
	ofstream pvslStats;
	ofstream pvsrStats;

	if (printStats)
	{
		char str[64]; 
		
		sprintf(str, "tmp/vsp_heur_sum-%04d.log", leaves);
		sumStats.open(str);
		sprintf(str, "tmp/vsp_heur_pvsl-%04d.log", leaves);
		pvslStats.open(str);
		sprintf(str, "tmp/vsp_heur_pvsr-%04d.log", leaves);
		pvsrStats.open(str);
	}

#endif


	for (ci = mLocalSubdivisionCandidates->begin(); ci != ci_end; ++ ci)
	{
		// compute changes to front and back pvs
		EvalHeuristics(*ci, pvsl, pvsr);

		// Note: sufficient to compare size of bounding boxes of front and back side?
		if (((*ci).value >= minBand) && ((*ci).value <= maxBand))
		{
			float currentPos;
			
			// HACK: current positition is BETWEEN visibility events
			if (0 && ((ci + 1) != ci_end))
				currentPos = ((*ci).value + (*(ci + 1)).value) * 0.5f;
			else
				currentPos = (*ci).value;			

			sum = pvsl * ((*ci).value - minBox) + pvsr * (maxBox - (*ci).value);
			
#ifdef GTP_DEBUG
			if (printStats)
			{
				sumStats
					<< "#Position\n" << currentPos << endl
					<< "#Sum\n" << sum * volRatio << endl
					<< "#Pvs\n" << pvsl + pvsr << endl;

				pvslStats
					<< "#Position\n" << currentPos << endl
					<< "#Pvsl\n" << pvsl << endl;

				pvsrStats
					<< "#Position\n" << currentPos << endl
					<< "#Pvsr\n" << pvsr << endl;
			}
#endif

			if (sum < minSum)
			{
				splitPlaneFound = true;

				minSum = sum;
				position = currentPos;
				
				pvsBack = pvsl;
				pvsFront = pvsr;
			}
		}
	}
	
	/////////	
	//-- compute cost

	const float pOverall = sizeBox;
	const float pBack = position - minBox;
	const float pFront = maxBox - position;

	const float penaltyOld = pvsCost;
    const float penaltyFront = pvsFront;
	const float penaltyBack = pvsBack;
	
	const float oldRenderCost = penaltyOld * pOverall + Limits::Small;
	const float newRenderCost = penaltyFront * pFront + penaltyBack * pBack;

#if VISUALIZE_SPLIT
	
	const int leaves = mVspStats.Leaves();
	const bool showViz = (leaves >= 3000) && (leaves % 497 == 0);

	if (showViz)
	{
		//cout << "========================== exporting split ===================" << endl;

		Exporter *exporter;

		char str[64]; sprintf(str, "vc-%04d_%d.wrl", leaves, axis);

		exporter = Exporter::GetExporter(str);

		Material m;
		m.mDiffuseColor.r = 1.0f;
		m.mDiffuseColor.g = 0.0f;
		m.mDiffuseColor.b = 0.0f;

		exporter->SetForcedMaterial(m);
		exporter->SetWireframe();

		exporter->ExportBox(tData.mBoundingBox);

		exporter->ResetForcedMaterial();

		RayInfoContainer::const_iterator rit, rit_end = tData.mRays->end();

		exporter->SetFilled();
	
		m.mDiffuseColor.r = 0.0f;
		m.mDiffuseColor.g = 1.0f;
		m.mDiffuseColor.b = 0.0f;

		exporter->SetForcedMaterial(m);

		// create unique ids for pvs heuristics
		Intersectable::NewMail(3);

		float a = 0, b = 0, c = 0;

		// this is the main ray classification loop!
		for(rit = tData.mRays->begin(); rit != rit_end; ++ rit)
		{
			VssRay *ray = (*rit).mRay;

			// determine the side of this ray with respect to the plane
			float t;
			const int side = (*rit).ComputeRayIntersection(axis, position, t);
	
			UpdateContributionsToPvs(*ray, true, side, a, b, c);
			UpdateContributionsToPvs(*ray, false, side, a, b, c);
		}


		for (rit = tData.mRays->begin(); rit != rit_end; ++ rit)
		{
			RayInfo rayInf = *rit;
			
			// export objects
			VssRay *ray = rayInf.mRay;	
			BvhLeaf *leaf = mBvHierarchy->GetLeaf(ray->mTerminationObject);

			//left and right
			if (leaf->Mailed(2))
			{
				m.mDiffuseColor.r = 0.0f;
				m.mDiffuseColor.g = 1.0f;
				m.mDiffuseColor.b = 1.0f;
			}
			else if (leaf->Mailed(1)) // only right
			{
				m.mDiffuseColor.r = 0.0f;
				m.mDiffuseColor.g = 1.0f;
				m.mDiffuseColor.b = 0.0f;
			}
			else // left
			{
				m.mDiffuseColor.r = 0.0f;
				m.mDiffuseColor.g = 0.0f;
				m.mDiffuseColor.b = 1.0f;
			}

			exporter->SetForcedMaterial(m);
			exporter->ExportIntersectable(leaf);

		}

		m.mDiffuseColor.r = 1.0f;
		m.mDiffuseColor.g = 0.0f;
		m.mDiffuseColor.b = 1.0f;

		exporter->SetForcedMaterial(m);
		AxisAlignedBox3 vizPlaneBox;
		vizPlaneBox = tData.mBoundingBox;
	
		vizPlaneBox.SetMin(axis, position - 0.01);
		vizPlaneBox.SetMax(axis, position + 0.01);

		exporter->ExportBox(vizPlaneBox);

		delete exporter;
	}

#endif

	if (splitPlaneFound)
	{
		ratio = newRenderCost / oldRenderCost;
	}
	

#ifdef GTP_DEBUG
	Debug << "\n((((( eval local vsp cost )))))" << endl
		  << "back pvs: " << penaltyBack << " front pvs: " << penaltyFront << " total pvs: " << penaltyOld << endl 
		  << "back p: " << pBack * volRatio << " front p " << pFront * volRatio << " p: " << pOverall * volRatio << endl
		  << "old rc: " << oldRenderCost * volRatio << " new rc: " << newRenderCost * volRatio << endl
		  << "render cost decrease: " << oldRenderCost * volRatio - newRenderCost * volRatio << endl;
#endif

	return ratio;
}


float VspTree::SelectSplitPlane(const VspTraversalData &tData,
								AxisAlignedPlane &plane,
								float &pFront,
								float &pBack)
{
	mSplitTimer.Entry();

	float nPosition[3];
	float nCostRatio[3];
	float nProbFront[3];
	float nProbBack[3];

	// create bounding box of node geometry
	AxisAlignedBox3 box = tData.mBoundingBox;
		
	int sAxis = 0;
	int bestAxis = -1;

	// do we use some kind of specialised "fixed" axis?
    const bool useSpecialAxis = 
		mOnlyDrivingAxis || mCirculatingAxis;
	
	// get subset of rays
	RayInfoContainer randomRays;
	randomRays.reserve(min(mMaxTests, (int)tData.mRays->size()));

	RayInfoContainer *usedRays;

	if (mMaxTests < (int)tData.mRays->size())
	{
		GetRayInfoSets(*tData.mRays, mMaxTests, randomRays);
		usedRays = &randomRays;
	}
	else
	{
		usedRays = tData.mRays;
	}

	if (mCirculatingAxis)
	{
		int parentAxis = 0;
		VspNode *parent = tData.mNode->GetParent();

		if (parent)
		{
			parentAxis = static_cast<VspInterior *>(parent)->GetAxis();
		}

		sAxis = (parentAxis + 1) % 3;
	}
	else if (mOnlyDrivingAxis)
	{
		sAxis = box.Size().DrivingAxis();
	}
	
	for (int axis = 0; axis < 3; ++ axis)
	{
		if (!useSpecialAxis || (axis == sAxis))
		{
			if (mUseCostHeuristics)
			{
				//-- place split plane using heuristics
				nCostRatio[axis] =
					EvalLocalCostHeuristics(tData,
											box,
											axis,
											nPosition[axis],
											*usedRays);			
			}
			else
			{
				//-- split plane position is spatial median				
				nPosition[axis] = (box.Min()[axis] + box.Max()[axis]) * 0.5f;
				nCostRatio[axis] = EvalLocalSplitCost(tData,
													  box,
													  axis,
													  nPosition[axis],
													  nProbFront[axis], 
													  nProbBack[axis]);
			}
						
			if (bestAxis == -1)
			{
				bestAxis = axis;
			}
			else if (nCostRatio[axis] < nCostRatio[bestAxis])
			{
				bestAxis = axis;
			} 
		}
	}

	/////////////////////////
	//-- assign values of best split

	plane.mAxis = bestAxis;
	// best split plane position
	plane.mPosition = nPosition[bestAxis];

	pFront = nProbFront[bestAxis];
	pBack = nProbBack[bestAxis];

	mSplitTimer.Exit();

	return nCostRatio[bestAxis];
}


float VspTree::EvalRenderCostDecrease(VspSubdivisionCandidate &sc,
									  float &normalizedOldRenderCost,
									  const SplitData &sData) const
{
	const float viewSpaceVol = mBoundingBox.GetVolume();
	const VspTraversalData &tData = sc.mParentData;
	
	const AxisAlignedPlane &candidatePlane = sc.mSplitPlane;
	const float avgRaysPerObject = sc.GetAvgRaysPerObject();


	AxisAlignedBox3 frontBox;
	AxisAlignedBox3 backBox;

	tData.mBoundingBox.Split(candidatePlane.mAxis, 
							 candidatePlane.mPosition, 
							 frontBox, 
							 backBox);

    // probability that view point lies in back / front node
	float pOverall = tData.mProbability;
	float pFront = frontBox.GetVolume();
	float pBack = pOverall - pFront;


	///////////////////
	//-- evaluate render cost heuristics

	// ratio of how render cost changed since last evaluation
	const float oldRenderCostRatio = (tData.mRenderCost > 0)? (sData.mTotalRenderCost / tData.mRenderCost) : 1;
	const float penaltyOld = tData.mCorrectedRenderCost * oldRenderCostRatio;

	sc.mCorrectedFrontRenderCost = mHierarchyManager->EvalCorrectedPvs(sData.mFrontRenderCost, penaltyOld, avgRaysPerObject);
	sc.mCorrectedBackRenderCost = mHierarchyManager->EvalCorrectedPvs(sData.mBackRenderCost, penaltyOld, avgRaysPerObject);

	sc.mFrontRenderCost = sData.mFrontRenderCost;
	sc.mBackRenderCost = sData.mBackRenderCost;

	const float oldRenderCost = penaltyOld * pOverall;
	const float newRenderCost = sc.mCorrectedFrontRenderCost * pFront + sc.mCorrectedBackRenderCost * pBack;


	// the normalized old render cost is needed for the priority
	normalizedOldRenderCost = oldRenderCost / viewSpaceVol;

	// the render cost decrase for this split
	const float renderCostDecrease = (oldRenderCost - newRenderCost) / viewSpaceVol;

#if GTP_DEBUG
	Debug << "old: " << normalizedOldRenderCost << " new: " << newRenderCost / viewSpaceVol 
		  << " corr: " << tData.mCorrectedRenderCost << " ratio: " << oldRenderCostRatio << endl;
#endif

	return renderCostDecrease;
}


float VspTree::EvalLocalSplitCost(const VspTraversalData &tData,
								  const AxisAlignedBox3 &box,
								  const int axis,
								  const float &position,
								  float &pFront,
								  float &pBack) const
{
	float pvsTotal = 0;
	float pvsFront = 0;
	float pvsBack = 0;
	
	// create unique ids for pvs heuristics
	Intersectable::NewMail(3);
	KdLeaf::NewMail(3);

	RayInfoContainer::const_iterator rit, rit_end = tData.mRays->end();

	// this is the main ray classification loop!
	for(rit = tData.mRays->begin(); rit != rit_end; ++ rit)
	{
		VssRay *ray = (*rit).mRay;

		// determine the side of this ray with respect to the plane
		float t;
		const int side = (*rit).ComputeRayIntersection(axis, position, t);
	
		UpdateContributionsToPvs(*ray, true, side, pvsFront, pvsBack, pvsTotal);
		UpdateContributionsToPvs(*ray, false, side, pvsFront, pvsBack, pvsTotal);
	}

	//////////////
	//-- evaluate cost heuristics

	float pOverall = tData.mProbability;

	// we use spatial mid split => simplified computation
	pBack = pFront = pOverall * 0.5f;
	
	const float newCost = pvsBack * pBack + pvsFront * pFront;
	const float oldCost = (float)pvsTotal * pOverall + Limits::Small;
	
#ifdef GTPGTP_DEBUG
	Debug << "axis: " << axis << " " << pvsSize << " " << pvsBack << " " << pvsFront << endl;
	Debug << "p: " << pFront << " " << pBack << " " << pOverall << endl;
#endif

	return  (mCtDivCi + newCost) / oldCost;
}


void VspTree::UpdateContributionsToPvs(Intersectable *obj,
									   const int cf,
									   float &frontPvs,
									   float &backPvs,
									   float &totalPvs) const
{
	if (!obj) return;

	// object in none of the pvss => new object
	if (!obj->Mailed() && !obj->Mailed(1) && !obj->Mailed(2))
	{
		totalPvs += 1.0f;
	}

	// QUESTION matt: is it safe to assume that 
	// the object belongs to no pvs in this case?
	//if (cf == Ray::COINCIDENT) return;
	if (cf >= 0) // front pvs
	{
		if (!obj->Mailed() && !obj->Mailed(2))
		{
			frontPvs += 1.0f;
		
			// already in back pvs => in both pvss
			if (obj->Mailed(1))
				obj->Mail(2);
			else
				obj->Mail();
		}
	}

	if (cf <= 0) // back pvs
	{
		if (!obj->Mailed(1) && !obj->Mailed(2))
		{
			backPvs += 1.0f;
		
			// already in front pvs => in both pvss
			if (obj->Mailed())
				obj->Mail(2);
			else
				obj->Mail(1);
		}
	}
}


void VspTree::UpdateContributionsToPvs(BvhLeaf *leaf,
									   const int cf,
									   float &frontPvs,
									   float &backPvs,
									   float &totalPvs,
									   const bool countEntries) const
{
	if (!leaf) return;

	const float renderCost = countEntries ? 1 : 
		(float)leaf->mObjects.size();
	
	// leaf in no pvs => new
	if (!leaf->Mailed() && !leaf->Mailed(1) && !leaf->Mailed(2))
	{
		totalPvs += renderCost;
	}

	if (cf >= 0) // front pvs
	{
		if (!leaf->Mailed() && !leaf->Mailed(2))
		{
			frontPvs += renderCost;
	
			// already in back pvs => in both pvss
			if (leaf->Mailed(1))
				leaf->Mail(2);
			else
				leaf->Mail();
		}
	}

	if (cf <= 0) // back pvs
	{
		if (!leaf->Mailed(1) && !leaf->Mailed(2))
		{
			backPvs += renderCost;
		
			// already in front pvs => in both pvss
			if (leaf->Mailed())
				leaf->Mail(2);
			else
				leaf->Mail(1);
		}
	}
}


void VspTree::UpdateContributionsToPvs(BvhLeaf *leaf,
									   const int cf,
									   SplitData &sData) const
{
	const int triSize = (int)leaf->mObjects.size();
	
	// object in no pvs => count as new
	if (!leaf->Mailed() && !leaf->Mailed(1) && !leaf->Mailed(2))
	{
		sData.mTotalTriangles += (int)triSize;
		++ sData.mTotalObjects;
	}

	if (cf >= 0) // front pvs
	{
		if (!leaf->Mailed() && !leaf->Mailed(2))
		{
			sData.mFrontTriangles += triSize;
			++ sData.mFrontObjects;

			// already in back pvs => in both pvss
			if (leaf->Mailed(1))
				leaf->Mail(2);
			else
				leaf->Mail();
		}
	}

	if (cf <= 0) // back pvs
	{
		if (!leaf->Mailed(1) && !leaf->Mailed(2))
		{
			sData.mBackTriangles += triSize;
			++ sData.mBackObjects;

			// already in front pvs => in both pvss
			if (leaf->Mailed())
				leaf->Mail(2);
			else
				leaf->Mail(1);
		}
	}
}


void VspTree::UpdateContributionsToPvs(KdLeaf *leaf,
									   const int cf,
									   float &frontPvs,
									   float &backPvs,
									   float &totalPvs) const
{
	if (!leaf) return;

	// the objects which are referenced in this and only this leaf
	const int contri = (int)(leaf->mObjects.size() - leaf->mMultipleObjects.size());
	
	// newly found leaf
	if (!leaf->Mailed() && !leaf->Mailed(1) && !leaf->Mailed(2))
	{
		totalPvs += contri;
	}

	// recursivly update contributions of yet unclassified objects
	ObjectContainer::const_iterator oit, oit_end = leaf->mMultipleObjects.end();

	for (oit = leaf->mMultipleObjects.begin(); oit != oit_end; ++ oit)
	{	
		UpdateContributionsToPvs(*oit, cf, frontPvs, backPvs, totalPvs);
    }	
	
	if (cf >= 0) // front pvs
	{
		if (!leaf->Mailed() && !leaf->Mailed(2))
		{
			frontPvs += contri;
		
			// already in back pvs => in both pvss
			if (leaf->Mailed(1))
				leaf->Mail(2);
			else
				leaf->Mail();
		}
	}

	if (cf <= 0) // back pvs
	{
		if (!leaf->Mailed(1) && !leaf->Mailed(2))
		{
			backPvs += contri;
		
			// already in front pvs => in both pvss
			if (leaf->Mailed())
				leaf->Mail(2);
			else
				leaf->Mail(1);
		}
	}
}


void VspTree::CollectLeaves(vector<VspLeaf *> &leaves, 
							const bool onlyUnmailed,
							const int maxPvsSize) const
{
	stack<VspNode *> nodeStack;
	nodeStack.push(mRoot);

	while (!nodeStack.empty())
	{
		VspNode *node = nodeStack.top();
		nodeStack.pop();
		
		if (node->IsLeaf())
		{
			// test if this leaf is in valid view space
			VspLeaf *leaf = static_cast<VspLeaf *>(node);

			if (leaf->TreeValid() && 
				(!onlyUnmailed || !leaf->Mailed()) &&
				((maxPvsSize < 0) || (leaf->GetViewCell()->GetPvs().EvalPvsCost() <= maxPvsSize)))
			{
				leaves.push_back(leaf);
			}
		}
		else
		{
			VspInterior *interior = static_cast<VspInterior *>(node);

			nodeStack.push(interior->GetBack());
			nodeStack.push(interior->GetFront());
		}
	}
}


AxisAlignedBox3 VspTree::GetBoundingBox() const
{
	return mBoundingBox;
}


VspNode *VspTree::GetRoot() const
{
	return mRoot;
}


void VspTree::EvaluateLeafStats(const VspTraversalData &data)
{
	// the node became a leaf -> evaluate stats for leafs
	VspLeaf *leaf = static_cast<VspLeaf *>(data.mNode);

	if (data.mPvs > mVspStats.maxPvs)
	{
		mVspStats.maxPvs = (int)data.mPvs;
	}

	mVspStats.pvs += (int)data.mPvs;

	if (data.mDepth < mVspStats.minDepth)
	{
		mVspStats.minDepth = data.mDepth;
	}
	
	if (data.mDepth >= mTermMaxDepth)
	{
        ++ mVspStats.maxDepthNodes;
	}

	// accumulate rays to compute rays /  leaf
	mVspStats.rayRefs += (int)data.mRays->size();

	if (data.mPvs < mTermMinPvs)
		++ mVspStats.minPvsNodes;

	if ((int)data.mRays->size() < mTermMinRays)
		++ mVspStats.minRaysNodes;

	if (data.GetAvgRayContribution() > mTermMaxRayContribution)
		++ mVspStats.maxRayContribNodes;

	if (data.mProbability <= mTermMinProbability)
		++ mVspStats.minProbabilityNodes;
	
	// accumulate depth to compute average depth
	mVspStats.accumDepth += data.mDepth;

	++ mCreatedViewCells;

#ifdef GTP_DEBUG
	Debug << "BSP stats: "
		  << "Depth: " << data.mDepth << " (max: " << mTermMaxDepth << "), "
		  << "PVS: " << data.mPvs << " (min: " << mTermMinPvs << "), "
		  << "#rays: " << (int)data.mRays->size() << " (max: " << mTermMinRays << "), "
		  << "#pvs: " << leaf->GetViewCell()->GetPvs().EvalPvsCost() << "), "
		  << "#avg ray contrib (pvs): " << (float)data.mPvs / (float)data.mRays->size() << endl;
#endif
}


void VspTree::CollectViewCells(ViewCellContainer &viewCells, bool onlyValid) const
{
	ViewCell::NewMail();
	CollectViewCells(mRoot, onlyValid, viewCells, true);
}


void VspTree::CollectRays(VssRayContainer &rays)
{
	vector<VspLeaf *> leaves;
	CollectLeaves(leaves);

	vector<VspLeaf *>::const_iterator lit, lit_end = leaves.end();

	for (lit = leaves.begin(); lit != lit_end; ++ lit)
	{
		VspLeaf *leaf = *lit;
		VssRayContainer::const_iterator rit, rit_end = leaf->mVssRays.end();

		for (rit = leaf->mVssRays.begin(); rit != rit_end; ++ rit)
			rays.push_back(*rit);
	}
}


void VspTree::SetViewCellsManager(ViewCellsManager *vcm)
{
	mViewCellsManager = vcm;
}


void VspTree::ValidateTree()
{
	if (!mRoot)
		return;

	mVspStats.invalidLeaves = 0;
	stack<VspNode *> nodeStack;

	nodeStack.push(mRoot);

	while (!nodeStack.empty()) 
	{
		VspNode *node = nodeStack.top();
		nodeStack.pop();
		
		if (node->IsLeaf())
		{
			VspLeaf *leaf = static_cast<VspLeaf *>(node);

			if (!leaf->GetViewCell()->GetValid())
				++ mVspStats.invalidLeaves;

			// validity flags don't match => repair
			if (leaf->GetViewCell()->GetValid() != leaf->TreeValid())
			{
				leaf->SetTreeValid(leaf->GetViewCell()->GetValid());
				PropagateUpValidity(leaf);
			}
		}
		else
		{
			VspInterior *interior = static_cast<VspInterior *>(node);
		
			nodeStack.push(interior->GetFront());
			nodeStack.push(interior->GetBack());
		}
	}

	Debug << "invalid leaves: " << mVspStats.invalidLeaves << endl;
}



void VspTree::CollectViewCells(VspNode *root,
							   bool onlyValid,
							   ViewCellContainer &viewCells,
							   bool onlyUnmailed) const
{
	if (!root)
		return;

	stack<VspNode *> nodeStack;
	nodeStack.push(root);
	
	while (!nodeStack.empty()) 
	{
		VspNode *node = nodeStack.top();
		nodeStack.pop();
		
		if (node->IsLeaf())
		{
			if (!onlyValid || node->TreeValid())
			{
				ViewCellLeaf *leafVc = static_cast<VspLeaf *>(node)->GetViewCell();

				ViewCell *viewCell = mViewCellsTree->GetActiveViewCell(leafVc);
						
				if (!onlyUnmailed || !viewCell->Mailed()) 
				{
					viewCell->Mail();
					viewCells.push_back(viewCell);
				}
			}
		}
		else
		{
			VspInterior *interior = static_cast<VspInterior *>(node);
		
			nodeStack.push(interior->GetFront());
			nodeStack.push(interior->GetBack());
		}
	}
}


int VspTree::FindNeighbors(VspLeaf *n,
						   vector<VspLeaf *> &neighbors,
						   const bool onlyUnmailed) const
{
	stack<VspNode *> nodeStack;
	nodeStack.push(mRoot);

	const AxisAlignedBox3 box = GetBoundingBox(n);

	while (!nodeStack.empty())
	{
		VspNode *node = nodeStack.top();
		nodeStack.pop();

		if (node->IsLeaf())
		{
			VspLeaf *leaf = static_cast<VspLeaf *>(node);

			if (leaf != n && (!onlyUnmailed || !leaf->Mailed()))
				neighbors.push_back(leaf);
		}
		else
		{
			VspInterior *interior = static_cast<VspInterior *>(node);
			
			if (interior->GetPosition() > box.Max(interior->GetAxis()))
				nodeStack.push(interior->GetBack());
			else
			{
				if (interior->GetPosition() < box.Min(interior->GetAxis()))
					nodeStack.push(interior->GetFront());
				else
				{
					// random decision
					nodeStack.push(interior->GetBack());
					nodeStack.push(interior->GetFront());
				}
			}
		}
	}

	return (int)neighbors.size();
}


// Find random neighbor which was not mailed
VspLeaf *VspTree::GetRandomLeaf(const Plane3 &plane)
{
	stack<VspNode *> nodeStack;
	nodeStack.push(mRoot);
  
	int mask = rand();
  
	while (!nodeStack.empty()) 
	{
		VspNode *node = nodeStack.top();
		
		nodeStack.pop();
		
		if (node->IsLeaf()) 
		{
			return static_cast<VspLeaf *>(node);
		} 
		else 
		{
			VspInterior *interior = static_cast<VspInterior *>(node);
			VspNode *next;
			
			if (GetBoundingBox(interior->GetBack()).Side(plane) < 0)
			{
				next = interior->GetFront();
			}
            else
			{
				if (GetBoundingBox(interior->GetFront()).Side(plane) < 0)
				{
					next = interior->GetBack();
				}
				else 
				{
					// random decision
					if (mask & 1)
						next = interior->GetBack();
					else
						next = interior->GetFront();
					mask = mask >> 1;
				}
			}
			
			nodeStack.push(next);
		}
	}
  
	return NULL;
}


VspLeaf *VspTree::GetRandomLeaf(const bool onlyUnmailed)
{
	stack<VspNode *> nodeStack;

	nodeStack.push(mRoot);

	int mask = rand();

	while (!nodeStack.empty())
	{
		VspNode *node = nodeStack.top();
		nodeStack.pop();

		if (node->IsLeaf())
		{
			if ( (!onlyUnmailed || !node->Mailed()) )
				return static_cast<VspLeaf *>(node);
		}
		else
		{
			VspInterior *interior = static_cast<VspInterior *>(node);

			// random decision
			if (mask & 1)
				nodeStack.push(interior->GetBack());
			else
				nodeStack.push(interior->GetFront());

			mask = mask >> 1;
		}
	}

	return NULL;
}


float VspTree::EvalPvsCost(const RayInfoContainer &rays) const
{
	float pvsCost = 0;
	
	Intersectable::NewMail();
	KdNode::NewMail();

	RayInfoContainer::const_iterator rit, rit_end = rays.end();

	for (rit = rays.begin(); rit != rays.end(); ++ rit)
	{
		VssRay *ray = (*rit).mRay;
		
		pvsCost += EvalContributionToPvs(*ray, true);

#if COUNT_ORIGIN_OBJECTS
		pvsCost += EvalContributionToPvs(*ray, false);
#endif
	}
	
	return pvsCost;
}


int VspTree::EvalPvsEntriesContribution(const VssRay &ray, 
										const bool isTermination) const

{

#if HACK_PERFORMANCE
	Intersectable *obj = isTermination ? ray.mTerminationObject : ray.mOriginObject; 

	if (!obj) return 0;
	BvhLeaf *bvhleaf = mBvHierarchy->GetLeaf(obj);

	if (!bvhleaf->Mailed())
	{
		bvhleaf->Mail();
		return 1;
	}

#else

	Intersectable *obj; 
	static Vector3 pt;
	KdNode *node;

	ray.GetSampleData(isTermination, pt, &obj, &node);
	
	if (!obj) return 0;

	switch(mHierarchyManager->GetObjectSpaceSubdivisionType())
	{
	case HierarchyManager::NO_OBJ_SUBDIV:
		{
			if (!obj->Mailed())
			{
				obj->Mail();
				return 1;
			}

			return 0;
		}

	case HierarchyManager::KD_BASED_OBJ_SUBDIV:
		{
			KdLeaf *leaf = mHierarchyManager->mOspTree->GetLeaf(pt, node);
			if (!leaf->Mailed())
			{
				leaf->Mail();
				return 1;
			}

			return 0;
		}
	case HierarchyManager::BV_BASED_OBJ_SUBDIV:
		{
			BvhLeaf *bvhleaf = mBvHierarchy->GetLeaf(obj);

			if (!bvhleaf->Mailed())
			{
				bvhleaf->Mail();
				return 1;
			}

			return 0;
		}
	default:
		break;
	}
#endif
	return 0;
}


int VspTree::EvalPvsEntriesSize(const RayInfoContainer &rays) const
{
	int pvsSize = 0;

	Intersectable::NewMail();
	KdNode::NewMail();

	RayInfoContainer::const_iterator rit, rit_end = rays.end();

	for (rit = rays.begin(); rit != rays.end(); ++ rit)
	{
		VssRay *ray = (*rit).mRay;

		pvsSize += EvalPvsEntriesContribution(*ray, true);

#if COUNT_ORIGIN_OBJECTS
		pvsSize += EvalPvsEntriesContribution(*ray, false);
#endif
	}

	return pvsSize;
}


float VspTree::GetEpsilon() const
{
	return mEpsilon;
}


int VspTree::CastLineSegment(const Vector3 &origin,
							 const Vector3 &termination,
                             ViewCellContainer &viewcells,
							 const bool useMailboxing)
{
	int hits = 0;

	float mint = 0.0f, maxt = 1.0f;
	const Vector3 dir = termination - origin;

	stack<LineTraversalData> tStack;

	Vector3 entp = origin;
	Vector3 extp = termination;

	VspNode *node = mRoot;
	VspNode *farChild;

	float position;
	int axis;

	while (1)
	{
		if (!node->IsLeaf())
		{
			VspInterior *in = static_cast<VspInterior *>(node);
			position = in->GetPosition();
			axis = in->GetAxis();

			if (entp[axis] <= position)
			{
				if (extp[axis] <= position)
				{
					node = in->GetBack();
					// cases N1,N2,N3,P5,Z2,Z3
					continue;
				} 
				else
				{
					// case N4
					node = in->GetBack();
					farChild = in->GetFront();
				}
			}
			else
			{
				if (position <= extp[axis])
				{
					node = in->GetFront();
					// cases P1,P2,P3,N5,Z1
					continue;
				}
				else
				{
					node = in->GetFront();
					farChild = in->GetBack();
					// case P4
				}
			}

 			// $$ modification 3.5.2004 - hints from Kamil Ghais
			// case N4 or P4
			const float tdist = (position - origin[axis]) / dir[axis];
			tStack.push(LineTraversalData(farChild, extp, maxt)); //TODO

			// new exit point for near child
			extp = origin + dir * tdist;
			maxt = tdist;
		}
		else
		{
			// compute intersection with all objects in this leaf
			VspLeaf *leaf = static_cast<VspLeaf *>(node);
			ViewCell *viewCell;

			if (0)
				viewCell = mViewCellsTree->GetActiveViewCell(leaf->GetViewCell());
			else
				viewCell = leaf->GetViewCell();

			// don't have to mail if each view cell belongs to exactly one leaf
			if (!useMailboxing || !viewCell->Mailed())
			{
				if (useMailboxing)
					viewCell->Mail();

				viewcells.push_back(viewCell);
				++ hits;
			}

			// get the next node from the stack
			if (tStack.empty())
				break;

			// set new entry point for far child
			entp = extp;
			mint = maxt;
			
			LineTraversalData &s  = tStack.top();
			node = s.mNode;
			extp = s.mExitPoint;
			maxt = s.mMaxT;

			tStack.pop();
		}
	}

	return hits;
}


int VspTree::TraverseRayPacket(RayPacket &rp, const bool useMailboxing)
{
#ifdef USE_SSE
	int hits = 0;

	// reciprocal of ray directions
	float one = 1.0f;
	float zero = 0.0f;

	__m128 one4 = _mm_load1_ps(&one);
	__m128 zero4 = _mm_load1_ps(&zero);

	union { __m128 mask4; float mask[4];};
	union { __m128 mask4_2; float mask_2[4];};

	mask4 = one4;

	__declspec(align(16)) __m128 entp4[] = {rp.mOriginX4, rp.mOriginY4, rp.mOriginZ4};
	__declspec(align(16)) __m128 extp4[] = {rp.mTerminationX4, rp.mTerminationY4, rp.mTerminationZ4};

	__m128 *origin = &rp.mOriginX4;
	__m128 *termination = &rp.mTerminationX4;

	// ray directions
	__m128 dirX4 = _mm_sub_ps(rp.mTerminationX4, rp.mOriginX4);
	__m128 dirY4 = _mm_sub_ps(rp.mTerminationY4, rp.mOriginY4);;
	__m128 dirZ4 = _mm_sub_ps(rp.mTerminationZ4, rp.mOriginZ4);;
	
	__m128 rdx4 = _mm_div_ps(one4, dirX4);
	__m128 rdy4 = _mm_div_ps(one4, dirY4);
	__m128 rdz4 = _mm_div_ps(one4, dirZ4);

	__m128 maxT4 = one4;
	__m128 minT4 = zero4;

	PacketTraversalStack tStack(1000);

	VspNode *node = mRoot;
	VspNode *farChild;

	float position;
	int axis;

	__m128 position4;

	while (1)
	{
		if (!node->IsLeaf())
		{
			VspInterior *in = static_cast<VspInterior *>(node);
		
			position = in->GetPosition();
			axis = in->GetAxis();

			position4 = _mm_load1_ps(&position);

			// are the entry points all in near leaf?
			if (_mm_movemask_ps(_mm_and_ps(_mm_cmple_ps(entp4[axis], position4), mask4)))
			{
				// are the exit points all in near leaf?
				if (_mm_movemask_ps(_mm_and_ps(_mm_cmple_ps(extp4[axis], position4), mask4)))
				{
					node = in->GetBack();
					// cases N1,N2,N3,P5,Z2,Z3
					continue;
				} 
				else // traverse both children
				{
					// case N4
					node = in->GetBack();
					farChild = in->GetFront();
				}
			}
			else
			{
				if (_mm_movemask_ps(_mm_and_ps(_mm_cmple_ps(position4, entp4[axis]), mask4)) &&
					_mm_movemask_ps(_mm_and_ps(_mm_cmple_ps(position4, extp4[axis]), mask4)))
				{
					node = in->GetFront();
					// cases P1,P2,P3,N5,Z1
					continue;
				}
				else // traverse both children
				{
					node = in->GetFront();
					farChild = in->GetBack();
					// case P4
				}
			}

 			// $$ modification 3.5.2004 - hints from Kamil Ghais
			// case N4 or P4
			const __m128 tDist4 = _mm_mul_ps(_mm_sub_ps(position4, origin[axis]), termination[axis]);
			
			mask4 = _mm_and_ps(_mm_cmplt_ps(tDist4, minT4), mask4);
            mask4_2 = _mm_and_ps(_mm_cmpgt_ps(tDist4, minT4), mask4);

			// push far child for further processing
			tStack.Push(PacketTraversalData(farChild, extp4[0], extp4[1], extp4[2], maxT4, mask4_2));

			// compute new exit point
			extp4[0] = _mm_add_ps(_mm_mul_ps(dirX4, tDist4), rp.mOriginX4);
			extp4[1] = _mm_add_ps(_mm_mul_ps(dirY4, tDist4), rp.mOriginY4);
			extp4[2] = _mm_add_ps(_mm_mul_ps(dirZ4, tDist4), rp.mOriginZ4);

			maxT4 = tDist4;
		}
		else
		{
			// compute intersection with all objects in this leaf
			VspLeaf *leaf = static_cast<VspLeaf *>(node);
			ViewCell *viewCell;

			viewCell = leaf->GetViewCell();

			// note: don't have to mail if each view 
			// cell belongs to exactly one leaf
			if (!useMailboxing || !viewCell->Mailed())
			{
				if (useMailboxing)
					viewCell->Mail();

				for (int i = 0; i < 4; ++i)
				{
					// push view cells for all valid rays
					if (mask[i]) 
						rp.mViewCells[i].push_back(viewCell);
				}

				++ hits;
			}

			// get the next node from the stack
			if (tStack.Empty())
				break;
			
			// use memcopy?
			entp4[0] = extp4[0];
			entp4[1] = extp4[1];
			entp4[2] = extp4[2];

			minT4 = maxT4;

			const PacketTraversalData &s = tStack.Top();
			node = s.mNode;
			
			extp4[0] = s.mExitPointX4;
			extp4[1] = s.mExitPointY4;
			extp4[2] = s.mExitPointZ4;

			maxT4 = s.mMaxT4;
			mask4 = s.mMask4;
			
			tStack.Pop();
		}
	}
	
	return hits;

#else

	return 0;
#endif
}


int VspTree::CastRay(Ray &ray)
{
	int hits = 0;

	stack<LineTraversalData> tStack;
	const Vector3 dir = ray.GetDir();

	float maxt, mint;

	if (!mBoundingBox.GetRaySegment(ray, mint, maxt))
		return 0;

	Intersectable::NewMail();
	ViewCell::NewMail();

	Vector3 entp = ray.Extrap(mint);
	Vector3 extp = ray.Extrap(maxt);

	const Vector3 origin = entp;

	VspNode *node = mRoot;
	VspNode *farChild = NULL;

	float position;
	int axis;

	while (1)
	{
		if (!node->IsLeaf())
		{
			VspInterior *in = static_cast<VspInterior *>(node);
			position = in->GetPosition();
			axis = in->GetAxis();

			if (entp[axis] <= position)
			{
				if (extp[axis] <= position)
				{
					node = in->GetBack();
					// cases N1,N2,N3,P5,Z2,Z3
					continue;
				} 
				else
				{
					// case N4
					node = in->GetBack();
					farChild = in->GetFront();
				}
			}
			else
			{
				if (position <= extp[axis])
				{
					node = in->GetFront();
					// cases P1,P2,P3,N5,Z1
					continue;
				}
				else
				{
					node = in->GetFront();
					farChild = in->GetBack();
					// case P4
				}
			}

 			// $$ modification 3.5.2004 - hints from Kamil Ghais
			// case N4 or P4
			const float tdist = (position - origin[axis]) / dir[axis];
			tStack.push(LineTraversalData(farChild, extp, maxt)); //TODO
			extp = origin + dir * tdist;
			maxt = tdist;
		}
		else
		{
			// compute intersection with all objects in this leaf
			VspLeaf *leaf = static_cast<VspLeaf *>(node);
			ViewCell *vc = leaf->GetViewCell();

			if (!vc->Mailed())
			{
				vc->Mail();
				// todo: add view cells to ray
				++ hits;
			}

			// get the next node from the stack
			if (tStack.empty())
				break;

			entp = extp;
			mint = maxt;
			
			LineTraversalData &s  = tStack.top();
			node = s.mNode;
			extp = s.mExitPoint;
			maxt = s.mMaxT;
			tStack.pop();
		}
	}

	return hits;
}


ViewCell *VspTree::GetViewCell(const Vector3 &point, const bool active)
{
	if (!mRoot)
		return NULL;

	stack<VspNode *> nodeStack;
	nodeStack.push(mRoot);
  
	ViewCellLeaf *viewcell = NULL;
  
	while (!nodeStack.empty())  
	{
		VspNode *node = nodeStack.top();
		nodeStack.pop();
	
		if (node->IsLeaf()) 
		{
			const AxisAlignedBox3 box = GetBoundingBox(static_cast<VspLeaf *>(node));

			if (!box.IsInside(point))
				Debug << "error, point " << point << " should be in view cell " << box << endl;
			
			viewcell = static_cast<VspLeaf *>(node)->GetViewCell();
			break;
		} 
		else 	
		{	
			VspInterior *interior = static_cast<VspInterior *>(node);
     
			// random decision
			if (interior->GetPosition() - point[interior->GetAxis()] < 0)
			{
				nodeStack.push(interior->GetFront());
			}
			else
			{
				nodeStack.push(interior->GetBack());
			}
		}
	}
  
	// return active or leaf view cell
	if (active)
	{
		return mViewCellsTree->GetActiveViewCell(viewcell);
	}
	else
	{
		return viewcell;
	}
}


bool VspTree::ViewPointValid(const Vector3 &viewPoint) const
{
	VspNode *node = mRoot;

	while (1)
	{
		// early exit
		if (node->TreeValid())
			return true;

		if (node->IsLeaf())
			return false;
			
		VspInterior *in = static_cast<VspInterior *>(node);
					
		if (in->GetPosition() - viewPoint[in->GetAxis()] <= 0) 
		{
			node = in->GetBack();
		} 
		else
		{
			node = in->GetFront();
		}
	}

	// should never come here
	return false;
}


void VspTree::PropagateUpValidity(VspNode *node)
{
	const bool isValid = node->TreeValid();

	// propagative up invalid flag until only invalid nodes exist over this node
	if (!isValid)
	{
		while (!node->IsRoot() && node->GetParent()->TreeValid())
		{
			node = node->GetParent();
			node->SetTreeValid(false);
		}
	}
	else
	{
		// propagative up valid flag until one of the subtrees is invalid
		while (!node->IsRoot() && !node->TreeValid())
		{
            node = node->GetParent();
			VspInterior *interior = static_cast<VspInterior *>(node);
			
			// the parent is valid iff both leaves are valid
			node->SetTreeValid(interior->GetBack()->TreeValid() && 
							   interior->GetFront()->TreeValid());
		}
	}
}


bool VspTree::Export(OUT_STREAM &stream)
{
	ExportNode(mRoot, stream);

	return true;
}


void VspTree::ExportNode(VspNode *node, OUT_STREAM &stream)
{
	if (node->IsLeaf())
	{
		VspLeaf *leaf = static_cast<VspLeaf *>(node);
		ViewCell *viewCell = mViewCellsTree->GetActiveViewCell(leaf->GetViewCell());

		int id = -1;
		if (viewCell != mOutOfBoundsCell)
			id = viewCell->GetId();

		stream << "<Leaf viewCellId=\"" << id << "\" />" << endl;
	}
	else
	{	
		VspInterior *interior = static_cast<VspInterior *>(node);
	
		AxisAlignedPlane plane = interior->GetPlane();
		stream << "<Interior plane=\"" << plane.mPosition << " " 
			   << plane.mAxis << "\">" << endl;

		ExportNode(interior->GetBack(), stream);
		ExportNode(interior->GetFront(), stream);

		stream << "</Interior>" << endl;
	}
}


int VspTree::SplitRays(const AxisAlignedPlane &plane,
					   RayInfoContainer &rays,
					   RayInfoContainer &frontRays,
					   RayInfoContainer &backRays) const
{
	int splits = 0;

	RayInfoContainer::const_iterator rit, rit_end = rays.end();

	for (rit = rays.begin(); rit != rit_end; ++ rit)
	{
		RayInfo bRay = *rit;
		
		VssRay *ray = bRay.mRay;
		float t;

		// get classification and receive new t
		// test if start point behind or in front of plane
		const int side = bRay.ComputeRayIntersection(plane.mAxis, plane.mPosition, t);
			
		if (side == 0)
		{
			++ splits;

			if (ray->HasPosDir(plane.mAxis))
			{
				backRays.push_back(RayInfo(ray, bRay.GetMinT(), t));
				frontRays.push_back(RayInfo(ray, t, bRay.GetMaxT()));
			}
			else
			{
				frontRays.push_back(RayInfo(ray, bRay.GetMinT(), t));
				backRays.push_back(RayInfo(ray, t, bRay.GetMaxT()));
			}
		}
		else if (side == 1)
		{
			frontRays.push_back(bRay);
		}
		else
		{
			backRays.push_back(bRay);
		}
	}

	return splits;
}


AxisAlignedBox3 VspTree::GetBoundingBox(VspNode *node) const
{
	if (!node->GetParent())
		return mBoundingBox;

	if (!node->IsLeaf())
	{
		return (static_cast<VspInterior *>(node))->GetBoundingBox();		
	}

	VspInterior *parent = static_cast<VspInterior *>(node->GetParent());

	AxisAlignedBox3 box(parent->GetBoundingBox());

	if (parent->GetFront() == node)
		box.SetMin(parent->GetAxis(), parent->GetPosition());
    else
		box.SetMax(parent->GetAxis(), parent->GetPosition());

	return box;
}


int VspTree::ComputeBoxIntersections(const AxisAlignedBox3 &box,
									 ViewCellContainer &viewCells) const
{
	stack<VspNode *> nodeStack;
 
	ViewCell::NewMail();

	nodeStack.push(mRoot);
	
	while (!nodeStack.empty()) 
	{
		VspNode *node = nodeStack.top();
		nodeStack.pop();

		const AxisAlignedBox3 bbox = GetBoundingBox(node);

		if (Overlap(bbox, box)) 
		{
			if (node->IsLeaf())
			{
				VspLeaf *leaf = static_cast<VspLeaf *>(node);

				if (!leaf->GetViewCell()->Mailed() && leaf->TreeValid())
				{
					leaf->GetViewCell()->Mail();
					viewCells.push_back(leaf->GetViewCell());
				}
			}
			else 
			{
				VspInterior *interior = static_cast<VspInterior *>(node);

				VspNode *first = interior->GetFront();
				VspNode *second = interior->GetBack();

				nodeStack.push(first);
				nodeStack.push(second);
			}
		}
		// default: cull
	}

	return (int)viewCells.size();
}


void VspTree::PreprocessRays(const VssRayContainer &sampleRays,
							 RayInfoContainer &rays)
{
	VssRayContainer::const_iterator rit, rit_end = sampleRays.end();

	long startTime = GetTime();

	cout << "storing rays ... ";

	Intersectable::NewMail();

	////////////////
	//-- store rays and objects
	
	VssRay *lastVssRay = NULL;

	for (rit = sampleRays.begin(); rit != rit_end; ++ rit)
	{
		VssRay *ray = *rit;

		// filter out double rays (last ray the same as this ray
		if (
			!lastVssRay ||
			!(ray->mOrigin == lastVssRay->mTermination) ||
			!(ray->mTermination == lastVssRay->mOrigin)) 
		{
			lastVssRay = ray;

			float minT, maxT;
			static Ray hray;

			hray.Init(*ray);
		
			// TODO: not very efficient to implictly cast between rays types
			if (GetBoundingBox().GetRaySegment(hray, minT, maxT))
			{
				const float len = ray->Length();

				if (len) // ray not degenerate 
				{	//len = Limits::Small;
					rays.push_back(RayInfo(ray, minT / len, maxT / len));
				}
			}
		}
		else
		{
			//cout << "r";
			// store object only for one ray
			//lastVssRay->mOriginObject = ray->mTerminationObject;
		}
	}

	cout << "finished in " << TimeDiff(startTime, GetTime()) * 1e-3 << " secs" << endl;
}


void VspTree::GetViewCells(const VssRay &ray, ViewCellContainer &viewCells)
{
	mViewCellsTimer.Entry();
	
	static Ray hray;
	hray.Init(ray);
	
	float tmin = 0, tmax = 1.0;

	if (!mBoundingBox.GetRaySegment(hray, tmin, tmax) || (tmin > tmax))
		return;

	const Vector3 origin = hray.Extrap(tmin);
	const Vector3 termination = hray.Extrap(tmax);

	// view cells were not precomputed and are extracted on the fly
	// don't mail because the same view cells can be found for different rays
	CastLineSegment(origin, termination, viewCells, false);

	mViewCellsTimer.Exit();
}


void VspTree::Initialise(const VssRayContainer &rays,
						 AxisAlignedBox3 *forcedBoundingBox,
						 const ObjectContainer &objects)
{
	ComputeBoundingBox(rays, forcedBoundingBox);

	VspLeaf *leaf = new VspLeaf();
	mRoot = leaf;

	VspViewCell *viewCell = new VspViewCell();
    leaf->SetViewCell(viewCell);

	viewCell->SetTrianglesInPvs((float)objects.size());
	viewCell->SetEntriesInPvs(1);

	// set view cell values
	viewCell->mLeaves.push_back(leaf);
	viewCell->SetVolume(mBoundingBox.GetVolume());

	leaf->mProbability = mBoundingBox.GetVolume();
}


void VspTree::PrepareConstruction(SplitQueue &tQueue,
								  const VssRayContainer &sampleRays,
								  RayInfoContainer &rays)
{	
	mVspStats.Reset();
	mVspStats.Start();
	mVspStats.nodes = 1;

	// store pointer to this tree
	VspSubdivisionCandidate::sVspTree = this;
	
	mBvHierarchy = mHierarchyManager->mBvHierarchy;

	// initialise termination criteria
	mTermMinProbability *= mBoundingBox.GetVolume();
	
	// get clipped rays
	PreprocessRays(sampleRays, rays);

	/// collect pvs from rays
	const float pvsCost = EvalPvsCost(rays);
	
	// root and bounding box were already constructed
	VspLeaf *leaf = static_cast<VspLeaf *>(mRoot);

	//////////
	//-- prepare view space partition

	const float prop = mBoundingBox.GetVolume();
	
	// first vsp traversal data
	VspTraversalData vData(leaf, 0, &rays, pvsCost, prop, mBoundingBox);

	mTotalCost = vData.mCorrectedRenderCost = vData.mRenderCost = pvsCost;
	mPvsEntries = EvalPvsEntriesSize(rays);
	vData.mCorrectedPvs = vData.mPvs = (float)mPvsEntries;

	//////////////
	//-- create the first split candidate

	VspSubdivisionCandidate *splitCandidate = new VspSubdivisionCandidate(vData);
    EvalSubdivisionCandidate(*splitCandidate);
	leaf->SetSubdivisionCandidate(splitCandidate);

	EvalSubdivisionStats(*splitCandidate);

	tQueue.Push(splitCandidate);
}


void VspTree::AddCandidateToDirtyList(const VssRay &ray, 
									  const bool isTermination,
									  vector<SubdivisionCandidate *> &dirtyList,
									  const bool onlyUnmailed) const

{
#if HACK_PERFORMANCE
	Intersectable *obj = isTermination ? ray.mTerminationObject : ray.mOriginObject; 

	if (!obj) return;

	BvhLeaf *leaf = mBvHierarchy->GetLeaf(obj);
	SubdivisionCandidate *candidate;

	if (!leaf->Mailed())
	{
		leaf->Mail();
		candidate = leaf->GetSubdivisionCandidate();
	}
	else
	{
		candidate = NULL;
	}
#else
	SubdivisionCandidate *candidate = NULL;

	Intersectable *obj;
	Vector3 pt;
	KdNode *node;

	ray.GetSampleData(isTermination, pt, &obj, &node);
	
	if (!obj) return;

	switch (mHierarchyManager->GetObjectSpaceSubdivisionType())
	{
	case HierarchyManager::KD_BASED_OBJ_SUBDIV:
		{
			KdLeaf *leaf = mHierarchyManager->mOspTree->GetLeaf(pt, node);

			if (!leaf->Mailed())
			{
				leaf->Mail();
				candidate = leaf->mSubdivisionCandidate;
			}
			break;
		}
	case HierarchyManager::BV_BASED_OBJ_SUBDIV:
		{
			BvhLeaf *leaf = mBvHierarchy->GetLeaf(obj);

			if (!leaf->Mailed())
			{
				leaf->Mail();
				candidate = leaf->GetSubdivisionCandidate();
			}
			break;
		}
	default:
		cerr << "collectdirtycandidates not implemented yet" << endl;
		candidate = NULL;
		break;
	}
#endif

	// is this leaf still a split candidate?
	if (candidate && (!onlyUnmailed || !candidate->Mailed()))
	{
		candidate->Mail();
		candidate->SetDirty(true);
		dirtyList.push_back(candidate);
	}
}


void VspTree::CollectDirtyCandidates(VspSubdivisionCandidate *sc, 
									 vector<SubdivisionCandidate *> &dirtyList,
									 const bool onlyUnmailed)
{
	VspTraversalData &tData = sc->mParentData;
	VspLeaf *node = tData.mNode;
	
	KdLeaf::NewMail();
	Intersectable::NewMail();
	
	RayInfoContainer::const_iterator rit, rit_end = tData.mRays->end();

	// add all nodes seen by the rays
	for (rit = tData.mRays->begin(); rit != rit_end; ++ rit)
	{
		VssRay *ray = (*rit).mRay;
		
		AddCandidateToDirtyList(*ray, true, dirtyList, onlyUnmailed);

#if COUNT_ORIGIN_OBJECTS
        AddCandidateToDirtyList(*ray, false, dirtyList, onlyUnmailed);
#endif
	}
}


float VspTree::EvalMaxEventContribution(const VssRay &ray, 
									  const bool isTermination) const
{
#if HACK_PERFORMANCE

	Intersectable *obj = isTermination ? ray.mTerminationObject : ray.mOriginObject; 

	if (!obj) return 0.0f;

	BvhLeaf *leaf = mBvHierarchy->GetLeaf(obj);
			
	// simple render cost evaluation
	if (-- leaf->mCounter == 0)
		return (float)leaf->mObjects.size();
	else
		return 0.0f;
#else

	Intersectable *obj;
	Vector3 pt;
	KdNode *node;

	ray.GetSampleData(isTermination, pt, &obj, &node);

	if (!obj) return 0.0f;

	float pvs = 0.0f;

	switch (mHierarchyManager->GetObjectSpaceSubdivisionType())
	{
	case HierarchyManager::NO_OBJ_SUBDIV:
		{
			if (-- obj->mCounter == 0)
				++ pvs;
			break;
		}
	case HierarchyManager::KD_BASED_OBJ_SUBDIV:
		{
			KdLeaf *leaf = mHierarchyManager->mOspTree->GetLeaf(pt, node);

			// add contributions of the kd nodes
			pvs += EvalMaxEventContribution(leaf);
			break;
		}
	case HierarchyManager::BV_BASED_OBJ_SUBDIV:
		{
			BvhLeaf *leaf = mBvHierarchy->GetLeaf(obj);
			
			// simple render cost evaluation
			if (-- leaf->mCounter == 0)
				pvs += BvHierarchy::EvalAbsCost(leaf->mObjects);

			break;
		}
	default:
		break;
	}
	return pvs;

#endif
}


float VspTree::PrepareHeuristics(const VssRay &ray, const bool isTermination)
{
	
#if HACK_PERFORMANCE
	
	Intersectable *obj = isTermination ? ray.mTerminationObject : ray.mOriginObject; 

	if (!obj) return 0.0f;

	BvhLeaf *leaf = mBvHierarchy->GetLeaf(obj);

	if (!leaf->Mailed())
	{
		leaf->Mail();
		leaf->mCounter = 1;
		return (float)leaf->mObjects.size();
	}
	else
	{
		++ leaf->mCounter;	
		return 0.0f;
	}

#else

	float pvsSize = 0;

	Intersectable *obj;
	Vector3 pt;
	KdNode *node;

	ray.GetSampleData(isTermination, pt, &obj, &node);

	if (!obj) return 0.0f;

	switch (mHierarchyManager->GetObjectSpaceSubdivisionType())
	{
	case HierarchyManager::NO_OBJ_SUBDIV:
		{
			if (!obj->Mailed())
			{
				obj->Mail();
				obj->mCounter = 0;
				++ pvsSize;
			}

			++ obj->mCounter;	
			break;
		}
	case HierarchyManager::KD_BASED_OBJ_SUBDIV:
		{
			KdLeaf *leaf = mHierarchyManager->mOspTree->GetLeaf(pt, node);
			pvsSize += PrepareHeuristics(leaf);	
			break;
		}
	case HierarchyManager::BV_BASED_OBJ_SUBDIV:
		{
			BvhLeaf *leaf = mBvHierarchy->GetLeaf(obj);

			if (!leaf->Mailed())
			{
				leaf->Mail();
				leaf->mCounter = 0;
				pvsSize += BvHierarchy::EvalAbsCost(leaf->mObjects);
			}

			++ leaf->mCounter;	
			break;
		}
	default:
		break;
	}
	return pvsSize;
#endif
}


float VspTree::EvalMinEventContribution(const VssRay &ray, 
										const bool isTermination) const
{
#if HACK_PERFORMANCE

	Intersectable *obj = isTermination ? ray.mTerminationObject : ray.mOriginObject; 

	if (!obj) return 0.0f;

	BvhLeaf *leaf = mBvHierarchy->GetLeaf(obj);
	
	if (!leaf->Mailed())
	{
		leaf->Mail();
		return (float)leaf->mObjects.size();
	}
	else
	{
		return 0.0f;
	}
#else

	Intersectable *obj;
	static Vector3 pt;
	KdNode *node;

	ray.GetSampleData(isTermination, pt, &obj, &node);

	if (!obj) return 0;

	float pvs = 0.0f;

	switch (mHierarchyManager->GetObjectSpaceSubdivisionType())
	{
	case HierarchyManager::NO_OBJ_SUBDIV:
		{
			if (!obj->Mailed())
			{
				obj->Mail();
				++ pvs;
			}
			break;
		}
	case HierarchyManager::KD_BASED_OBJ_SUBDIV:
		{
			KdLeaf *leaf = mHierarchyManager->mOspTree->GetLeaf(pt, node);
			// add contributions of the kd nodes
			pvs += EvalMinEventContribution(leaf);				
			break;
		}
	case HierarchyManager::BV_BASED_OBJ_SUBDIV:
		{
			BvhLeaf *leaf = mBvHierarchy->GetLeaf(obj);
			if (!leaf->Mailed())
			{
				leaf->Mail();
				pvs += BvHierarchy->EvalAbsCost(leaf->mObjects);
			}
			break;
		}
	default:
		break;
	}
	return pvs;

#endif
}


void VspTree::UpdateContributionsToPvs(const VssRay &ray,
									   const bool isTermination,
									   const int cf,
									   float &frontPvs,
									   float &backPvs,
									   float &totalPvs) const
{
#if HACK_PERFORMANCE

	BvhLeaf *leaf = mBvHierarchy->GetLeaf(ray.mTerminationObject);
	UpdateContributionsToPvs(leaf, cf, frontPvs, backPvs, totalPvs, false);

#else

	Intersectable *obj;
	static Vector3 pt;
	KdNode *node;

	ray.GetSampleData(isTermination, pt, &obj, &node);

	if (!obj) return;

	switch (mHierarchyManager->GetObjectSpaceSubdivisionType())
	{
		case HierarchyManager::NO_OBJ_SUBDIV:
		{
			// find front and back pvs for origing and termination object
			UpdateContributionsToPvs(obj, cf, frontPvs, backPvs, totalPvs);
			break;
		}
		case HierarchyManager::KD_BASED_OBJ_SUBDIV:
		{
			KdLeaf *leaf = mHierarchyManager->mOspTree->GetLeaf(pt, node);
			UpdateContributionsToPvs(leaf, cf, frontPvs, backPvs, totalPvs);
			break;
		}
		case HierarchyManager::BV_BASED_OBJ_SUBDIV:
		{
			BvhLeaf *leaf = mBvHierarchy->GetLeaf(obj);
			UpdateContributionsToPvs(leaf, cf, frontPvs, backPvs, totalPvs, false);
			break;
		}
	}
#endif
}


void VspTree::UpdatePvsEntriesContribution(const VssRay &ray,
										   const bool isTermination,
										   const int cf,
										   float &pvsFront,
										   float &pvsBack,
										   float &totalPvs) const
{	
#if HACK_PERFORMANCE

	BvhLeaf *leaf = mBvHierarchy->GetLeaf(ray.mTerminationObject);
	UpdateContributionsToPvs(leaf, cf, pvsFront, pvsBack, totalPvs, true);

#else

	Intersectable *obj;
	static Vector3 pt;
	KdNode *node;

	ray.GetSampleData(isTermination, pt, &obj, &node);
	if (!obj) return;

	switch (mHierarchyManager->GetObjectSpaceSubdivisionType())
	{
	case HierarchyManager::KD_BASED_OBJ_SUBDIV:
		// TODO
		break;
	case HierarchyManager::BV_BASED_OBJ_SUBDIV:
		{
			BvhLeaf *leaf = mBvHierarchy->GetLeaf(obj);
			UpdateContributionsToPvs(leaf, cf, pvsFront, pvsBack, totalPvs, true);
			break;
		}
	default:
		{
			UpdateContributionsToPvs(obj, cf, pvsFront, pvsBack, totalPvs);	
			break;
		}
	}
#endif
}


float VspTree::EvalContributionToPvs(const VssRay &ray, const bool isTermination) const
{

#if HACK_PERFORMANCE

	Intersectable *obj = isTermination ? ray.mTerminationObject : ray.mOriginObject; 

	if (!obj) return 0;

	BvhLeaf *leaf = mBvHierarchy->GetLeaf(obj);

	if (!leaf->Mailed())
	{
		leaf->Mail();
		return (float)leaf->mObjects.size();
	}
	else
	{
		return 0.0f;
	}

#else
	
	Intersectable *obj; 
	static Vector3 pt;
	KdNode *node;

	ray.GetSampleData(isTermination, pt, &obj, &node);

	if (!obj) return 0;

	float pvs = 0.0f;

	switch (mHierarchyManager->GetObjectSpaceSubdivisionType())
	{
	case HierarchyManager::NO_OBJ_SUBDIV:
		{
			if (!obj->Mailed())
			{
				obj->Mail();
				++ pvs;
			}
			break;
		}
	case HierarchyManager::KD_BASED_OBJ_SUBDIV:
		{
			KdLeaf *leaf = mHierarchyManager->mOspTree->GetLeaf(pt, node);
			pvs += EvalContributionToPvs(leaf);
			break;
		}
	case HierarchyManager::BV_BASED_OBJ_SUBDIV:
		{
			BvhLeaf *bvhleaf = mBvHierarchy->GetLeaf(obj);

			if (!bvhleaf->Mailed())
			{
				bvhleaf->Mail();
				pvs += BvHierarchy::EvalAbsCost(bvhleaf->mObjects);
			}
			break;
		}
	default:
		break;
	}
	return pvs;

#endif
}


float VspTree::EvalContributionToPvs(KdLeaf *leaf) const
{
	if (leaf->Mailed()) // leaf already mailed
		return 0;
	
	leaf->Mail();

	// this is the pvs which is uniquely part of this kd leaf
	float pvs = (float)(leaf->mObjects.size() - leaf->mMultipleObjects.size());

	ObjectContainer::const_iterator oit, oit_end = leaf->mMultipleObjects.end();

	for (oit = leaf->mMultipleObjects.begin(); oit != oit_end; ++ oit)
	{
		Intersectable *obj = *oit;
		if (!obj->Mailed())
		{
			obj->Mail();
			++ pvs;
		}
	}

	return pvs;
}


int VspTree::CompressObjects(VspLeaf *leaf)
{
	bool compressed = true;

	while (compressed)
	{
		BvhNode::NewMail(2);

		ObjectPvsIterator oit = leaf->GetViewCell()->GetPvs().GetIterator();
		vector<BvhNode *> parents;
		ObjectPvs newPvs;

		compressed = false;

		while (oit.HasMoreEntries())
		{
			BvhNode *obj = static_cast<BvhNode *>(oit.Next());

			if (!obj->IsRoot())
			{
				BvhNode *parent = obj->GetParent();

				if (!parent->Mailed())
				{
					if (parent->Mailed(1))
						cout << "error!!" << endl;
					parent->Mail();
				}
				else
				{
					// sibling was already found => 
					// this entry can be exchanged by the parent
					parents.push_back(parent);
					parent->Mail(1);
		
					compressed = true;
				}
			}
		}

		PvsData pvsData;
		oit = leaf->GetViewCell()->GetPvs().GetIterator();

		while (oit.HasMoreEntries())
		{
			BvhNode *obj = static_cast<BvhNode *>(oit.Next(pvsData));

			if (!obj->IsRoot())
			{
				BvhNode *parent = obj->GetParent();

				// add only entries that cannot be exchanged with the parent
				if (!parent->Mailed(1))
				{
					newPvs.AddSampleDirty(obj, pvsData.mSumPdf);
				}
			}
		}

		// add parents
		vector<BvhNode *>::const_iterator bit, bit_end = parents.end();

		for (bit = parents.begin(); bit != bit_end; ++ bit)
		{
			newPvs.AddSampleDirty(*bit, 1);
		}

		//cout << "size " << newPvs.GetSize() << endl;
		leaf->GetViewCell()->SetPvs(newPvs);
	}

	return leaf->GetViewCell()->GetPvs().GetSize();
}


int VspTree::CompressObjects()
{
	vector<VspLeaf *> leaves;
	CollectLeaves(leaves);

	int numEntries = 0;

	vector<VspLeaf *>::const_iterator lit, lit_end = leaves.end();

	for (lit = leaves.begin(); lit != lit_end; ++ lit)
	{
		numEntries += CompressObjects(*lit);
	}

	return numEntries;
}


void VspTree::EvalMinMaxDepth(int &minDepth, int &maxDepth)
{
	stack<pair<VspNode *, int> > nodeStack;
	nodeStack.push(pair<VspNode *, int>(mRoot, 0));

	minDepth = 999999;
	maxDepth = 0;

	while (!nodeStack.empty())
	{
		VspNode *node = nodeStack.top().first;
		const int depth = nodeStack.top().second;

		nodeStack.pop();
		
		if (node->IsLeaf())
		{
			// test if this leaf is in valid view space
			VspLeaf *leaf = static_cast<VspLeaf *>(node);

			if (depth < minDepth)
				minDepth = depth;

			if (depth > maxDepth)
				maxDepth = depth;
		}
		else
		{
			VspInterior *interior = static_cast<VspInterior *>(node);

			nodeStack.push(pair<VspNode *, int>(interior->GetBack(), depth + 1));
			nodeStack.push(pair<VspNode *, int>(interior->GetFront(), depth + 1));
		}
	}
}


//////////
// Binary export

void VspTree::ExportBinInterior(OUT_STREAM &stream, VspInterior *interior)
{
	int type = TYPE_INTERIOR;
	stream.write(reinterpret_cast<char *>(&type), sizeof(int));

	int axis = interior->GetAxis();
	float pos = interior->GetPosition();

	stream.write(reinterpret_cast<char *>(&axis), sizeof(int));
	stream.write(reinterpret_cast<char *>(&pos), sizeof(float));
}


void VspTree::ExportBinLeaf(OUT_STREAM &stream, VspLeaf *leaf)
{
	int type = TYPE_LEAF;
	stream.write(reinterpret_cast<char *>(&type), sizeof(int));
}


bool VspTree::ExportBinary(OUT_STREAM &stream)
{
	// export binary version of mesh
	queue<VspNode *> tStack;
	tStack.push(mRoot);

	while(!tStack.empty())
	{
		VspNode *node = tStack.front();
		tStack.pop();
		
		if (node->IsLeaf())
		{
			ExportBinLeaf(stream, static_cast<VspLeaf *>(node));
		}
		else
		{
			VspInterior *interior = static_cast<VspInterior *>(node);

			ExportBinInterior(stream, interior);
			
			tStack.push(interior->GetFront());
			tStack.push(interior->GetBack());
		}
	}

	return true;
}


//////////////////
// import binary

VspInterior *VspTree::ImportBinInterior(IN_STREAM  &stream, VspInterior *parent)
{
	AxisAlignedPlane plane;

	stream.read(reinterpret_cast<char *>(&plane.mAxis), sizeof(int));
	stream.read(reinterpret_cast<char *>(&plane.mPosition), sizeof(float));

	VspInterior *interior = new VspInterior(plane);

	return interior;
}


VspLeaf *VspTree::ImportBinLeaf(IN_STREAM &stream, 
								VspInterior *parent)
{
	int leafId = TYPE_LEAF;
	int objId = leafId;
	
	VspLeaf *leaf = new VspLeaf(parent);

	// pvs is loaded by view cell
	return leaf;
}


VspNode *VspTree::ImportNextNode(IN_STREAM &stream, 
								 VspInterior *parent)
{
	int nodeType;
	stream.read(reinterpret_cast<char *>(&nodeType), sizeof(int));

	if (nodeType == TYPE_LEAF)
	{
		//cerr << "l";
		return ImportBinLeaf(stream, static_cast<VspInterior *>(parent));
	}

	if (nodeType == TYPE_INTERIOR)
	{
		//cerr << "i";
		return ImportBinInterior(stream, static_cast<VspInterior *>(parent));
	}

	cerr << "error! loading vsp node failed!" << endl;
	return NULL;
}


/** Data used for tree traversal
*/
struct MyTraversalData
{  
public:

	MyTraversalData(VspNode *node, const int depth, const AxisAlignedBox3 &box):
	mNode(node), mDepth(depth), mBox(box)
	{}


	//////////////////////

	/// the current node
	VspNode *mNode;
	/// current depth
	int mDepth;
	/// the bounding box
	AxisAlignedBox3 mBox;
};


bool VspTree::ImportBinary(IN_STREAM &stream)
{
	// export binary version of mesh
	queue<MyTraversalData> tStack;

	// sort objects by their id
	//sort(objects.begin(), objects.end(), ilt);

	// hack: we make a new root
	DEL_PTR(mRoot);
	mRoot = ImportNextNode(stream, NULL);

	tStack.push(MyTraversalData(mRoot, 0, mBoundingBox));
	mVspStats.Reset();
	mVspStats.nodes = 1;

	while(!tStack.empty())
	{
		MyTraversalData tData = tStack.front();
		tStack.pop();

		VspNode *node = tData.mNode;

		if (!node->IsLeaf())
		{
			mVspStats.nodes += 2;

			VspInterior *interior = static_cast<VspInterior *>(node);
			interior->SetBoundingBox(tData.mBox);

            VspNode *front = ImportNextNode(stream, interior);
			VspNode *back = ImportNextNode(stream, interior);
	
			interior->SetupChildLinks(front, back);

			++ mVspStats.splits[interior->GetAxis()];

			// compute new bounding box
			AxisAlignedBox3 frontBox, backBox;
			
			tData.mBox.Split(interior->GetAxis(), 
                			 interior->GetPosition(), 
							 frontBox, 
							 backBox);

			tStack.push(MyTraversalData(front, tData.mDepth + 1, frontBox));			
			tStack.push(MyTraversalData(back, tData.mDepth + 1,  backBox));
		}
		else
		{
			//EvaluateLeafStats(tData);
			//cout << "l";
		}
	}

	return true;
}


void VspTree::TestOutput(const std::string &filename)
{
	ofstream str(filename.c_str());

	vector<VspLeaf *> vspLeaves;
	CollectLeaves(vspLeaves);

	str << "root: " << mBoundingBox << endl;

	vector<VspLeaf *>::const_iterator vit, vit_end = vspLeaves.end();

	for (vit = vspLeaves.begin(); vit != vit_end; ++ vit)
	{
		VspLeaf *leaf = *vit;
		ViewCell *vc = leaf->GetViewCell();

		str << "\nleaf: " << GetBoundingBox(leaf) << endl;

		ObjectContainer pvsObjects;

		ObjectPvsIterator pit = vc->GetPvs().GetIterator();

		while (pit.HasMoreEntries())
		{
			Intersectable *obj = pit.Next();
			pvsObjects.push_back(obj);
		}

		sort(pvsObjects.begin(), pvsObjects.end(), ilt);
		str << "pvs: " << pvsObjects.size() << endl;
		
		for (int i = 0; i < pvsObjects.size(); ++ i)
			str << pvsObjects[i]->GetId() << " ";
	}
}


}
